"""
    Diagnostic variables classes
    =============================

    Classes defining multiple scalar diagnostics (variables) of the model and used to analyze its outputs.

    Description of the classes
    --------------------------

    * :class:`VariablesDiagnostic`: General class to get and show the scalar variables of the models.
    * :class:`GeopotentialHeightDifferenceDiagnostic`: Class to compute and show the geopotential height difference
      between points of the model's domain.

"""
import warnings
import base64
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.animation as animation
from IPython.display import HTML, display
from ipywidgets import interactive

from qgs.diagnostics.base import Diagnostic

from qgs.diagnostics.misc import time_axis_label


class VariablesDiagnostic(Diagnostic):
    """General class to create multiple scalar diagnostics based on the variables of the model.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, variable_list, model_params, dimensional):

        Diagnostic.__init__(self, model_params, dimensional)
        self._variable_list = variable_list
        self._variable_labels = [self._model_params.latex_var_string[var] for var in self._variable_list]
        self._variable_units = [self._model_params.get_variable_units(var) for var in self._variable_list]
        self._plot_title = "Model's variables"

    def __len__(self):
        if self.diagnostic is not None:
            return self.diagnostic.shape[1]
        else:
            return None

    def _configure(self, **kwargs):
        pass

    def _get_diagnostic(self, dimensional):
        if dimensional:
            vr = self._model_params.variables_range
            for i, j in enumerate(self._variable_list):
                v = self._data[j].copy()
                if j < vr[0]:
                    v *= self._model_params.streamfunction_scaling
                if vr[0] <= j < vr[1]:
                    v *= self._model_params.temperature_scaling * 2
                if self._ocean:
                    if vr[1] <= j < vr[2]:
                        v *= self._model_params.streamfunction_scaling
                    if vr[2] <= j < vr[3]:
                        v *= self._model_params.temperature_scaling
                if self._ground:
                    if vr[1] <= j < vr[2]:
                        v *= self._model_params.temperature_scaling

                if i == 0:
                    self._diagnostic_data = v[np.newaxis, :]
                else:
                    self._diagnostic_data = np.vstack((self._diagnostic_data, v))
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = self._data[self._variable_list]
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data

    def plot(self, time_index=0, variables='all', style="timeserie", ax=None, figsize=(16, 9), plot_kwargs=None, **kwargs):
        """Plot the multiple scalar diagnostic provided.

        Parameters
        ----------
        time_index: int
            The time index of the data. Not used in this subclass.
        variables: str or list(int)
            List of the model variables to consider as diagnostics.
            Default to `all`, i.e. select all the variables of the model.
        style: str, optional
            The style of the plot. Can be:

            * `timeserie`: Plot all the selected variables as a function of time.
            * `2Dscatter`: Plot the first two selected variables on a 2D scatter plot.
            * `3Dscatter`: Plot the first three selected variables on a 3D scatter plot.

        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        plot_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.

        Returns
        -------
        ~matplotlib.axes.Axes
            An axes where the data were plotted.
        """

        if variables == 'all':
            variables = list(range(len(self._variable_list)))

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        if plot_kwargs is None:
            plot_kwargs = dict()

        if ax is None:
            fig = plt.figure(figsize=figsize)
            if '3D' in style:
                ax = fig.add_subplot(1, 1, 1, projection='3d')
            else:
                ax = fig.add_subplot(1, 1, 1)
        else:
            fig = ax.figure

        if style == 'timeserie':

            for var in variables:
                lab = '$' + self._variable_labels[var] + '$'
                if self.dimensional:
                    lab += r" [" + self._variable_units[var] + r"]"
                time = self._time * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time)
                ax.plot(time, self.diagnostic[var], label=lab, **plot_kwargs)
            ax.legend()

            if self.dimensional:
                ax.set_xlabel(time_axis_label + ' [' + self._model_params.time_unit + ']')
            else:
                ax.set_xlabel(time_axis_label + ' [timeunits]')

        elif style == "2Dscatter":
            ax.plot(self.diagnostic[variables[0]], self.diagnostic[variables[1]], marker='o', ls='', **plot_kwargs)

            # TODO: Still to test (and adapt to variables ranges)
            # if self._ocean:
            #     natm = self._model_params.nmod[0]
            #     ngoc = self._model_params.nmod[1]
            #     if 2*natm <= variables[0] < 2*natm+ngoc:
            #         tl = ax.get_xticklabels()
            #         labto, num = _tick_fmt(tl)
            #         ax.set_xticklabels(labto)
            #         til = ax.xaxis.get_major_ticks()
            #         for j in range(0, len(til), 1):
            #             til[j].label1.set_visible(False)
            #         til[num].label1.set_visible(True)
            #         til[0].label1.set_visible(True)

            lab = '$' + self._variable_labels[0] + '$'
            if self.dimensional:
                lab += r" [" + self._variable_units[0] + r"]"
            ax.set_xlabel(lab)
            lab = '$' + self._variable_labels[1] + '$'
            if self.dimensional:
                lab += r" [" + self._variable_units[1] + r"]"
            ax.set_ylabel(lab)

        elif style == "3Dscatter":
            ax.plot(self.diagnostic[variables[0]], self.diagnostic[variables[1]], self.diagnostic[variables[2]], marker='o', ls='', **plot_kwargs)

            lab = '$' + self._variable_labels[0] + '$'
            if self.dimensional:
                lab += r" [" + self._variable_units[0] + r"]"
            ax.set_xlabel(lab, labelpad=20.)
            lab = '$' + self._variable_labels[1] + '$'
            if self.dimensional:
                lab += r" [" + self._variable_units[1] + r"]"
            ax.set_ylabel(lab, labelpad=20.)
            lab = '$' + self._variable_labels[2] + '$'
            if self.dimensional:
                lab += r" [" + self._variable_units[2] + r"]"
            ax.set_zlabel(lab, labelpad=20.)

        else:
            warnings.warn('Provided style parameter ' + style + ' not supported ! Nothing to plot.')

        title = self._plot_title
        if self.dimensional:
            pass
        else:
            title += r" (in nondim units)"

        ax.set_title(title, pad=20.)

        return ax

    def movie(self, variables='all', output='html', filename='', style="2Dscatter", background=None, ax=None, figsize=(16, 9),
              plot_kwargs=None, anim_kwargs=None):
        """ Create and return a movie of the output of the `plot` method animated over time.
        Show a red dot moving and depicting the current value of the model's selected variables.

        Parameters
        ----------
        variables: str or list(int)
            List of the model variables to consider as diagnostics.
            Default to `all`, i.e. select all the variables of the model.
        output: str, optional
            Define the kind of movie being created. Can be:

            * `jshtml`: Generate an interactive HTML representation of the animation.
            * `html5`: Generate the movie as HTML5 code.
            * `html`: Output the movie as a HTML video tag.
            * `ihtml`: Output the interactive movie as a HTML video tag.
            * `save`: Save the movie in MP4 format (H264 codec).

            Default to `html`.
        filename: str, optional
            Filename (and path) where to save the movie. Needed if `output` is set to `save`.
        style: str, optional
            The style of the plot. Can be:

            * `timeserie`: Plot all the selected variables as a function of time.
            * `moving-timeserie`: Plot all the selected variables as a function of time. Draw the lines as the time evolves.
            * `2Dscatter`: Plot the first two selected variables on a 2D scatter plot.
            * `3Dscatter`: Plot the first three selected variables on a 3D scatter plot.

        background: VariablesDiagnostic, optional
            The variables diagnostic data used as background for the evolving red dot. If `None`, use the current
            VariablesDiagnostic instance. Default to `None`.
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot.
            If not provided, create a new one.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        plot_kwargs: dict, optional
            Arguments to pass to the background plot.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or HTML code or HTML tag
            The animation object or the HTML code or tag.
        """

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        anim = self._make_anim(variables, style, background, ax, figsize, True, plot_kwargs, anim_kwargs, False)

        if 'html' in output:

            if output == "jshtml" or output == 'ihtml':
                jshtml = anim.to_jshtml()
                if output == "jshtml":
                    return jshtml
                else:
                    return HTML(jshtml)
            else:
                html5 = anim.to_html5_video()
                if output == 'html5':
                    return html5
                else:
                    return HTML(html5)

        elif output == 'save':

            if not filename:
                warnings.warn('No filename provided to the method animate. Video not saved !\n Please provide a filename.')

            html = anim.to_html5_video()
            start_index = html.index('base64,')
            start_index += len('base64,')
            end_index = html.index('">', start_index)
            video = html[start_index: end_index]
            with open(filename, 'wb') as f:
                f.write(base64.b64decode(video))

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def animate(self, variables='all', output='animate', style="2Dscatter", background=None, ax=None, figsize=(16, 9), show_time=True,
                stride=1, plot_kwargs=None, anim_kwargs=None, show=True):
        """Return the output of the `plot` method animated over time.
        Show a red dot moving and depicting the current value of the model's selected variables.

        Parameters
        ----------
        variables: str or list(int)
            List of the model variables to consider as diagnostics.
            Default to `all`, i.e. select all the variables of the model.
        output: str, optional
            Define the kind of animation being created. Can be:

            * `animate`: Create and show a :class:`ipywidgets.widgets.interaction.interactive` widget. Works only in Jupyter notebooks.
            * `show`: Create and show an animation with the :mod:`matplotlib.animation` module. Works only in IPython or Python.

        style: str, optional
            The style of the plot. Can be:

            * `timeserie`: Plot all the selected variables as a function of time.
            * `moving-timeserie`: Plot all the selected variables as a function of time. Draw the lines as the time evolves.
            * `2Dscatter`: Plot the first two selected variables on a 2D scatter plot.
            * `3Dscatter`: Plot the first three selected variables on a 3D scatter plot.

        background: VariablesDiagnostic, optional
            The variables diagnostic data used as background for the evolving red dot. If `None`, use the current
            VariablesDiagnostic instance. Default to `None`.
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot.
            If not provided, create a new one.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        show_time: bool, optional
            Show the timestamp on the plot or not. Only valid for scatter plots.
            Default to `True`.
        stride: int, optional
            Specify the time step of the animation. Works only with `output` set to `animate`.
        plot_kwargs: dict, optional
            Arguments to pass to the background plot.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.
            Works only with `output` set to `show`.
        show: bool, optional
            Whether to plot or not the animation.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or ~IPython.display.DisplayHandle or callable
            The animation object or the callable to update the widget, depending on the value of the `output` and `show` parameters.
        """

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        if output == 'animate':

            if background is None:
                background = self

            if variables == 'all':
                variables = list(range(len(self._variable_list)))

            def update(time_index):
                if 'moving' not in style:
                    pack = background.plot(variables=variables, style=style, ax=ax, figsize=figsize, plot_kwargs=plot_kwargs)
                    if hasattr(pack, '__getitem__'):
                        axe = pack[0]
                    else:
                        axe = pack
                else:
                    if ax is None:
                        fig = plt.figure(figsize=figsize)
                        if '3D' in style:
                            axe = fig.add_subplot(1, 1, 1, projection='3d')
                        else:
                            axe = fig.add_subplot(1, 1, 1)
                    else:
                        axe = ax
                        fig = axe.figure

                if style == 'timeserie':

                    lines = ax.get_lines()
                    time = self._time[time_index:time_index+1] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time)
                    for i in range(len(variables)):
                        ax.plot(time, self.diagnostic[variables[i]][time_index:time_index+1], marker='o', linestyle='', color=lines[i].get_c())

                elif style == 'moving-timeserie':

                    ii = -1
                    lines = list()
                    var_min = 2.e168
                    var_max = -2.e168
                    end_time = -2.e168
                    start_time = 2.e168
                    for i in range(len(variables)):
                        var_min = min(var_min, np.min(self.diagnostic[i]))
                        var_max = max(var_max, np.max(self.diagnostic[i]))
                        end_time = max(end_time, self._time[ii] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time))
                        start_time = min(start_time, self._time[0] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time))
                        time = self._time[:time_index+1] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time)
                        lab = '$' + self._variable_labels[i] + '$'
                        if self.dimensional:
                            lab += r" [" + self._variable_units[i] + r"]"
                        line, = ax.plot(time, self.diagnostic[variables[i]][:time_index+1], label=lab)  # , **plot_kwargs
                        lines.append(line)
                    if var_min < 0.:
                        var_min *= 1.03
                    else:
                        var_min *= 0.97
                    if var_max > 0.:
                        var_max *= 1.03
                    else:
                        var_max *= 0.97
                    ax.set_xlim(start_time, end_time)
                    ax.set_ylim(var_min, var_max)
                    ax.legend()
                    if self.dimensional:
                        ax.set_xlabel(time_axis_label + ' [' + self._model_params.time_unit + ']')
                    else:
                        ax.set_xlabel(time_axis_label + ' [timeunits]')
                    title = self._plot_title
                    if self.dimensional:
                        pass
                    else:
                        title += r" (in nondim units)"
                    ax.set_title(title, pad=20.)

                elif style == '2Dscatter':
                    axe.plot(self.diagnostic[variables[0]][time_index:time_index + 1], self.diagnostic[variables[1]][time_index:time_index + 1], marker='o', linestyle='', color='r')
                elif style == '3Dscatter':
                    axe.plot(self.diagnostic[variables[0]][time_index:time_index + 1], self.diagnostic[variables[1]][time_index:time_index + 1], zs=self.diagnostic[variables[2]][time_index:time_index + 1],
                             marker='o', linestyle='', color='r')
                if show_time and 'scatter' in style:
                    if self.dimensional:
                        tt = " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[time_index]) + " " + self._model_params.time_unit
                    else:
                        tt = " at " + str(self._time[time_index]) + " timeunits"
                    if style == '2Dscatter':
                        axe.text(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=axe.transAxes)
                    elif style == '3Dscatter':
                        axe.text2D(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=axe.transAxes)
                if show:
                    plt.show()

            if show:
                plot = interactive(update, time_index=(0, len(self)-1, stride))
                anim = display(plot)
            else:
                return update

        elif output == 'show':

            anim = self._make_anim(variables, style, background, ax, figsize, show_time, plot_kwargs, anim_kwargs, True)
            if show:
                plt.show()

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def _init_anim(self, variables='all', style="2Dscatter", background=None, ax=None, figsize=(16, 9),
                   show_time_in_title=False, show_time=True, plot_kwargs=None, anim_kwargs=None):

        if variables == 'all':
            variables = list(range(len(self._variable_list)))

        if background is None:
            background = self

        if ax is None:
            fig = plt.figure(figsize=figsize)
            if '3D' in style:
                ax = fig.add_subplot(1, 1, 1, projection='3d')
            else:
                ax = fig.add_subplot(1, 1, 1)
        else:
            fig = ax.figure

        if 'moving' not in style:
            background.plot(variables=variables, style=style, ax=ax, figsize=figsize, plot_kwargs=plot_kwargs)

        if 'timeserie' in style:
            show_time = False

        if show_time and 'scatter' in style:
            if self.dimensional:
                tt = " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[0]) + " " + self._model_params.time_unit
            else:
                tt = " at " + str(self._time[0]) + " timeunits"
            if style == '2Dscatter':
                time_text = ax.text(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
            elif style == '3Dscatter':
                time_text = ax.text2D(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
        else:
            time_text = None

        if style == 'timeserie':
            xpoint = list()
            lines = ax.get_lines()
            for i in range(len(variables)):
                pp, = ax.plot([], [], marker=',', linestyle='', color=lines[i].get_c())
                xpoint.append(pp)
        elif style == 'moving-timeserie':
            if 'frames' in anim_kwargs:
                ii = anim_kwargs['frames']
            else:
                ii = -1
            lines = list()
            var_min = 2.e168
            var_max = -2.e168
            end_time = -2.e168
            start_time = 2.e168
            for i in range(len(variables)):
                var_min = min(var_min, np.min(self.diagnostic[i]))
                var_max = max(var_max, np.max(self.diagnostic[i]))
                end_time = max(end_time, self._time[ii] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time))
                start_time = min(start_time, self._time[0] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time))
                time = self._time[:1] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time)
                lab = '$' + self._variable_labels[i] + '$'
                if self.dimensional:
                    lab += r" [" + self._variable_units[i] + r"]"
                line, = ax.plot(time, self.diagnostic[variables[i]][:1], label=lab)  # , **plot_kwargs
                lines.append(line)
            if var_min < 0.:
                var_min *= 1.03
            else:
                var_min *= 0.97
            if var_max > 0.:
                var_max *= 1.03
            else:
                var_max *= 0.97
            ax.set_xlim(start_time, end_time)
            ax.set_ylim(var_min, var_max)
            ax.legend()
            if self.dimensional:
                ax.set_xlabel(time_axis_label + ' [' + self._model_params.time_unit + ']')
            else:
                ax.set_xlabel(time_axis_label + ' [timeunits]')
            title = self._plot_title
            if self.dimensional:
                pass
            else:
                title += r" (in nondim units)"

            ax.set_title(title, pad=20.)
            xpoint = None
        elif style == '2Dscatter':
            xpoint, = ax.plot([], [], marker=',', linestyle='', color='r')
            lines = None
        elif style == '3Dscatter':
            xpoint, = ax.plot([], [], zs=[], marker=',', linestyle='', color='r')
            lines = None
        else:
            warnings.warn('Provided style parameter ' + style + ' not supported ! Nothing to plot. Returning None.')
            return None

        fargs = (ax, show_time)

        kwargs = {'variables': variables, 'style': style, 'background': background,
                  'ax': ax, 'figsize': figsize,
                  'show_time': show_time, 'plot_kwargs': plot_kwargs,
                  'anim_kwargs': anim_kwargs,
                  'xpoint': xpoint,
                  'lines': lines,
                  'time_text': time_text}

        return fig, ax, fargs, kwargs

    def _make_update(self, variables='all', style="2Dscatter", background=None, ax=None, figsize=(16, 9),
                     show_time_in_title=False, show_time=True, plot_kwargs=None, anim_kwargs=None, time_text=None, lines=None, xpoint=None):

        if style == 'timeserie':
            def update(i, axe, show_t):
                if i == 1:
                    for xp in xpoint:
                        xp.set_marker('o')

                time = self._time[i:i+1] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time)
                for j, xp in enumerate(xpoint):
                    xp.set_data(time, self.diagnostic[variables[j]][i:i+1])
                return [axe]

        elif style == 'moving-timeserie':
            def update(i, axe, show_t):

                time = self._time[0:i+1] * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time)

                for j, line in enumerate(lines):
                    line.set_xdata(time)
                    line.set_ydata(self.diagnostic[variables[j]][0:i+1])

                return [axe]

        elif style == '2Dscatter':
            def update(i, axe, show_t):
                if i == 1:
                    xpoint.set_marker('o')
                xpoint.set_data(self.diagnostic[variables[0]][i:i+1], self.diagnostic[variables[1]][i:i+1])
                if show_t:
                    if self.dimensional:
                        tt = " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[i]) + " " + self._model_params.time_unit
                    else:
                        tt = " at " + str(self._time[i]) + " timeunits"
                    time_text.set_text(tt)
                return [axe]

        elif style == '3Dscatter':
            def update(i, axe, show_t):
                if i == 1:
                    xpoint.set_marker('o')
                xpoint.set_data(self.diagnostic[variables[0]][i:i+1], self.diagnostic[variables[1]][i:i+1])
                xpoint.set_3d_properties(self.diagnostic[variables[2]][i:i+1])
                if show_t:
                    if self.dimensional:
                        tt = " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[i]) + " " + self._model_params.time_unit
                    else:
                        tt = " at " + str(self._time[i]) + " timeunits"
                    time_text.set_text(tt)
                return [axe]
        else:
            warnings.warn('Provided style parameter ' + style + ' not supported ! Nothing to plot. Returning None.')
            return None

        return update

    def _make_anim(self, variables='all', style="2Dscatter", background=None, ax=None, figsize=(16, 9),
                   show_time=True, plot_kwargs=None, anim_kwargs=None, blit=True):

        fig, ax, fargs, kwargs = self._init_anim(variables, style, background, ax, figsize, show_time, plot_kwargs, anim_kwargs)

        update = self._make_update(**kwargs)

        if 'blit' in anim_kwargs:
            del anim_kwargs['blit']

        if anim_kwargs is not None:
            anim = animation.FuncAnimation(fig, update, fargs=fargs, blit=blit, **anim_kwargs)
        else:
            anim = animation.FuncAnimation(fig, update, fargs=fargs, blit=blit)

        return anim


class GeopotentialHeightDifferenceDiagnostic(VariablesDiagnostic):
    """Class to compute and show the geopotential height difference
    between points of the model's domain.

    Parameters
    ----------

    points_list: list(2-tuple(2-tuple(float)))
        List of couple of point (as 2-tuple of float) of which to compute the geopotential height difference.
    model_params: QgParams
        An instance of the model parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """
    def __init__(self, points_list, model_params, dimensional):

        self._point1 = list()
        self._point2 = list()
        self._func_points1 = None
        self._func_points2 = None
        variable_list = list(range(len(points_list)))

        VariablesDiagnostic.__init__(self, variable_list, model_params, dimensional)
        self._configure(points_list)
        self._plt = 'Geopotential height difference between points'
        self._plot_units = ' (in meters)'
        if self.dimensional:
            self._plot_title = self._plt + self._plot_units
        else:
            self._plot_title = self._plt

    def _configure(self, points_list, **kwargs):
        self.set_points(points_list)

    def set_points(self, points_list):
        """Set the couples of points of the domain of which to compute the geopotential height difference.

        Parameters
        ----------

        points_list: list(2-tuple(2-tuple(float)))
            List of couple of point (as 2-tuple of float) of which to compute the geopotential height difference.
        """
        self._point1 = list()
        self._point2 = list()
        for points in points_list:
            self._point1.append(points[0])
            self._point2.append(points[1])
        self._variable_labels = [ r'\mathrm{Points} \, (' + "{:.2f}".format(self._point1[i][0])+','
                                 + "{:.2f}".format(self._point1[i][1]) +r') \, \mathrm{and} \, (' + "{:.2f}".format(self._point2[i][0]) +
                                 ',' + "{:.2f}".format(self._point2[i][1])+')' for i in range(len(self._point1))]
        if self.dimensional:
            for i in range(len(self._variable_labels)):
                self._variable_labels[i] += r'\, \mathrm{in} \, '
        self._variable_units = len(self._point1) * ['m']
        self._compute_functions_value()

    def _compute_functions_value(self):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        self._func_points1 = list()
        self._func_points2 = list()

        basis = self._model_params.atmospheric_basis

        funcs_list = basis.num_functions(self._subs)

        for point in self._point1:
            self._func_points1.append(list())
            for func in funcs_list[offset:]:
                self._func_points1[-1].append(func(point[0], point[1]))

        for point in self._point2:
            self._func_points2.append(list())
            for func in funcs_list[offset:]:
                self._func_points2[-1].append(func(point[0], point[1]))

        self._func_points1 = np.array(self._func_points1)
        self._func_points2 = np.array(self._func_points2)

    def _get_diagnostic(self, dimensional):
        vr = self._model_params.variables_range
        for i in range(len(self._variable_list)):
            val1 = self._func_points1[i] @ self._data[:vr[0], ...]
            val2 = self._func_points2[i] @ self._data[:vr[0], ...]
            v = val1 - val2
            if i == 0:
                self._diagnostic_data = v[np.newaxis, :]
            else:
                self._diagnostic_data = np.vstack((self._diagnostic_data, v))

        if dimensional:
            self._diagnostic_data *= self._model_params.geopotential_scaling * self._model_params.streamfunction_scaling
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


if __name__ == '__main__':
    from qgs.params.params import QgParams
    from qgs.params.params import QgParams
    from qgs.integrators.integrator import RungeKuttaIntegrator
    from qgs.functions.tendencies import create_tendencies

    pars = QgParams()
    pars.set_atmospheric_channel_fourier_modes(2, 2)
    f, Df = create_tendencies(pars)
    integrator = RungeKuttaIntegrator()
    integrator.set_func(f)
    ic = np.random.rand(pars.ndim) * 0.1
    integrator.integrate(0., 200000., 0.1, ic=ic, write_steps=5)
    time, traj = integrator.get_trajectories()
    integrator.terminate()

    var_nondim = VariablesDiagnostic([10, 0, 14], pars, False)
    var_dim = VariablesDiagnostic([10, 0, 14], pars, True)

    var_nondim(time, traj)
    var_dim(time, traj)

    geo_nondim = GeopotentialHeightDifferenceDiagnostic([[[np.pi/pars.scale_params.n, np.pi/4], [np.pi/pars.scale_params.n, 3*np.pi/4]],
                                                         [[0, np.pi/4], [0, 3*np.pi/4]]],
                                                        pars, False)
    geo_nondim.set_data(time, traj)

    geo_dim = GeopotentialHeightDifferenceDiagnostic([[[np.pi/pars.scale_params.n, np.pi/4], [np.pi/pars.scale_params.n, 3*np.pi/4]],
                                                         [[0, np.pi/4], [0, 3*np.pi/4]]],
                                                        pars, True)
    geo_dim.set_data(time, traj)



"""
    Parameter module
    ================

    This module contains the basic parameter class to hold model's parameters values.
    It allows to manipulate dimensional and nondimensional parameter easily.

    Examples
    --------

    >>> from qgs.params.params import ScaleParams
    >>> from qgs.params.parameter import Parameter, ParametersArray
    >>> import numpy as np
    >>> # defining a scale object to help Parameter compute the nondimensionalization
    >>> sc = ScaleParams()
    >>> # creating a parameter initialized with a nondimensional value but returning a
    >>> # dimensional one when called
    >>> sigma = Parameter(0.2e0, input_dimensional=False, scale_object=sc,
    ...                   units='[m^2][s^-2][Pa^-2]',
    ...                   description="static stability of the atmosphere",
    ...                   return_dimensional=True)
    >>> sigma
    2.1581898457499433e-06
    >>> sigma.nondimensional_value
    0.2
    >>> sigma.return_dimensional
    True
    >>> # creating a parameter initialized with a dimensional value but returning a
    >>> # nondimensional one when called
    >>> sigma = Parameter(2.1581898457499433e-06, input_dimensional=True, scale_object=sc,
    ...                   units='[m^2][s^-2][Pa^-2]',
    ...                   description="static stability of the atmosphere",
    ...                   return_dimensional=False)
    >>> sigma
    0.2
    >>> sigma.dimensional_value
    2.1581898457499433e-06
    >>> sigma.return_dimensional
    False
    >>> # creating a parameters array initialized with a nondimensional values and returning
    >>> # nondimensional ones when called
    >>> s = ParametersArray(np.array([[0.1,0.2],[0.3,0.4]]), input_dimensional=False, scale_object=sc, units='[s^-1]',
    ...                     description="atmosphere bottom friction coefficient")
    >>> s
    ArrayParameters([[0.1, 0.2],
                     [0.3, 0.4]], dtype=object)
    >>> # dimensional values can also be retrieved with
    >>> s.dimensional_values
    array([[1.0320000000000001e-05, 2.0640000000000002e-05],
           [3.096e-05, 4.1280000000000005e-05]], dtype=object)
    >>> # you can also ask for the dimensional value of one particular value of the array
    >>> s[0,0]
    0.1
    >>> s[0,0].dimensional_value
    1.0320000000000001e-05

    Main class
    ----------
"""

import warnings
import numpy as np
from fractions import Fraction


# TODO: Automatize warnings and errors

class ScalingParameter(float):
    """Class of model's dimension parameter.

    Parameters
    ----------
    value: float
        Value of the parameter.
    units: str, optional
        The units of the provided value. Used to compute the conversion between dimensional and nondimensional
        value. Should be specified by joining atoms like `'[unit^power]'`, e.g '`[m^2][s^-2][Pa^-2]'`.
        Empty by default.
    description: str, optional
        String describing the parameter.
    dimensional: bool, optional
        Indicate if the value of the parameter is dimensional or not. Default to `True`.
    symbol: ~sympy.core.symbol.Symbol, optional
        A `Sympy`_ symbol to represent the parameter in symbolic expressions.
    symbolic_expression: ~sympy.core.expr.Expr, optional
        A `Sympy`_ expression to represent a relationship to other parameters.

    Notes
    -----
    Parameter is immutable. Once instantiated, it cannot be altered. To create a new parameter, one must
    re-instantiate it.

    .. _Sympy: https://www.sympy.org/
    """

    def __new__(cls, value, units="", description="", dimensional=False, symbol=None, symbolic_expression=None):

        f = float.__new__(cls, value)
        f._dimensional = dimensional
        f._units = units
        f._description = description
        f._symbol = symbol
        f._symbolic_expression = symbolic_expression

        return f

    @property
    def symbol(self):
        """~sympy.core.symbol.Symbol: Returns the symbol of the parameter."""
        return self._symbol

    @property
    def symbolic_expression(self):
        """~sympy.core.expr.Expr: Returns the symbolic expression of the parameter."""
        if self._symbolic_expression is None and self._symbol is not None:
            return self._symbol
        else:
            return self._symbolic_expression

    @property
    def dimensional(self):
        """bool: Indicate if the returned value is dimensional or not."""
        return self._dimensional

    @property
    def units(self):
        """str: The units of the dimensional value."""
        return self._units

    @property
    def description(self):
        """str: Description of the parameter."""
        return self._description

    def __add__(self, other):

        res = float(self) + other
        if isinstance(other, (Parameter, ScalingParameter)):
            if self.units != other.units:
                raise ArithmeticError("ScalingParameter class: Impossible to add two parameters with different units.")
            if self.symbolic_expression is None:
                if other.symbolic_expression is None:
                    if self.symbol is not None and other.symbol is not None:
                        expr = self.symbol + other.symbol
                    else:
                        expr = None
                    descr = self.description + " + " + other.description
                else:
                    if self.symbol is not None:
                        expr = self.symbol + (other.symbolic_expression)
                        descr = self.description + " + (" + other.description + ")"
                    else:
                        expr = None
                        descr = self.description + " + " + other.description
            else:
                if other.symbolic_expression is None:
                    if other.symbol is not None:
                        expr = (self.symbolic_expression) + other.symbol
                        descr = "(" + self.description + ") + " + other.description
                    else:
                        expr = None
                        descr = self.description + " + " + other.description
                else:
                    expr = (self.symbolic_expression) + (other.symbolic_expression)
                    descr = "(" + self.description + ") + (" + other.description + ")"

            if isinstance(other, Parameter):
                return Parameter(res, input_dimensional=other.input_dimensional,
                                 return_dimensional=other.return_dimensional, scale_object=other._scale_object,
                                 description=descr, units=self.units, symbol=None, symbolic_expression=expr)

            else:
                return ScalingParameter(res, description=descr, units=self.units, symbol=None, symbolic_expression=expr)
        else:
            try:
                if self.symbol is not None:
                    expr = self.symbol + other
                    descr = self.description + " + " + str(other)
                elif self.symbolic_expression is not None:
                    expr = (self.symbolic_expression) + other
                    descr = "(" + self.description + ") + " + str(other)
                else:
                    expr = None
                    descr = self.description + " + " + str(other)
                return ScalingParameter(res, description=descr, units=self.units, symbol=None, symbolic_expression=expr)
            except:
                return res

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):

        res = float(self) - other
        if isinstance(other, (Parameter, ScalingParameter)):
            if self.units != other.units:
                raise ArithmeticError("ScalingParameter class: Impossible to subtract two parameters with different units.")
            if self.symbolic_expression is None:
                if other.symbolic_expression is None:
                    if self.symbol is not None and other.symbol is not None:
                        expr = self.symbol - other.symbol
                    else:
                        expr = None
                    descr = self.description + " - " + other.description
                else:
                    if self.symbol is not None:
                        expr = self.symbol - (other.symbolic_expression)
                        descr = self.description + " - (" + other.description + ")"
                    else:
                        expr = None
                        descr = self.description + " - " + other.description
            else:
                if other.symbolic_expression is None:
                    if other.symbol is not None:
                        expr = (self.symbolic_expression) - other.symbol
                        descr = "(" + self.description + ") - " + other.description
                    else:
                        expr = None
                        descr = self.description + " - " + other.description
                else:
                    expr = (self.symbolic_expression) - (other.symbolic_expression)
                    descr = "(" + self.description + ") - (" + other.description + ")"

            if isinstance(other, Parameter):
                return Parameter(res, input_dimensional=other.input_dimensional,
                                 return_dimensional=other.return_dimensional, scale_object=other._scale_object,
                                 description=descr, units=self.units, symbol=None, symbolic_expression=expr)

            else:
                return ScalingParameter(res, description=descr, units=self.units, symbol=None, symbolic_expression=expr)
        else:
            try:
                if self.symbol is not None:
                    expr = self.symbol - other
                    descr = self.description + " - " + str(other)
                elif self.symbolic_expression is not None:
                    expr = (self.symbolic_expression) - other
                    descr = "(" + self.description + ") - " + str(other)
                else:
                    expr = None
                    descr = self.description + " - " + str(other)
                return ScalingParameter(res, description=descr, units=self.units, symbol=None, symbolic_expression=expr)
            except:
                return res

    def __rsub__(self, other):
        try:
            res = other - float(self)
            if self.symbol is not None:
                expr = other - self.symbol
                descr = str(other) + " - " + self.description
            elif self.symbolic_expression is not None:
                expr = other - (self.symbolic_expression)
                descr = str(other) + " - (" + self.description + ")"
            else:
                expr = None
                descr = str(other) + " - " + self.description
            return ScalingParameter(res, description=descr, units=self.units, symbol=None, symbolic_expression=expr)
        except:
            return res

    def __mul__(self, other):

        res = float(self) * other
        if isinstance(other, (Parameter, ScalingParameter)):
            units = _combine_units(self.units, other.units, '+')

            if self.symbolic_expression is None:
                if other.symbolic_expression is None:
                    if self.symbol is not None and other.symbol is not None:
                        expr = self.symbol * other.symbol
                    else:
                        expr = None
                    descr = self.description + " * " + other.description
                else:
                    if self.symbol is not None:
                        expr = self.symbol * (other.symbolic_expression)
                        descr = self.description + " * (" + other.description + ")"
                    else:
                        expr = None
                        descr = self.description + " * " + other.description
            else:
                if other.symbolic_expression is None:
                    if other.symbol is not None:
                        expr = (self.symbolic_expression) * other.symbol
                        descr = "(" + self.description + ") * " + other.description
                    else:
                        expr = None
                        descr = self.description + " * " + other.description
                else:
                    expr = (self.symbolic_expression) * (other.symbolic_expression)
                    descr = "(" + self.description + ") * (" + other.description + ")"

            if isinstance(other, Parameter):
                return Parameter(res, input_dimensional=other.input_dimensional, return_dimensional=other.return_dimensional,
                                 scale_object=other._scale_object, description=descr,
                                 units=units, symbol=None, symbolic_expression=expr)

            else:
                return ScalingParameter(res, description=descr, units=units, symbol=None, symbolic_expression=expr)
        else:
            try:
                if self.symbol is not None:
                    expr = self.symbol * other
                    descr = self.description + " * " + str(other)
                elif self.symbolic_expression is not None:
                    expr = (self.symbolic_expression) * other
                    descr = "(" + self.description + ") * " + str(other)
                else:
                    expr = None
                    descr = self.description + " * " + str(other)
                return ScalingParameter(res, description=descr, units=self.units, symbol=None, symbolic_expression=expr)
            except:
                return res

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):

        res = float(self) / other
        if isinstance(other, (ScalingParameter, Parameter)):
            units = _combine_units(self.units, other.units, '-')
            if self.symbolic_expression is None:
                if other.symbolic_expression is None:
                    if self.symbol is not None and other.symbol is not None:
                        expr = self.symbol / other.symbol
                    else:
                        expr = None
                    descr = self.description + " / " + other.description
                else:
                    if self.symbol is not None:
                        expr = self.symbol / (other.symbolic_expression)
                        descr = self.description + " / (" + other.description + ")"
                    else:
                        expr = None
                        descr = self.description + " / " + other.description
            else:
                if other.symbolic_expression is None:
                    if other.symbol is not None:
                        expr = (self.symbolic_expression) / other.symbol
                        descr = "(" + self.description + ") / " + other.description
                    else:
                        expr = None
                        descr = self.description + " / " + other.description
                else:
                    expr = (self.symbolic_expression) / (other.symbolic_expression)
                    descr = "(" + self.description + ") / (" + other.description + ")"

            if isinstance(other, Parameter):
                return Parameter(res, input_dimensional=other.input_dimensional, return_dimensional=other.return_dimensional,
                                 scale_object=other._scale_object, description=descr,
                                 units=units, symbol=None, symbolic_expression=expr)
            else:
                return ScalingParameter(res, description=descr, units=units, symbol=None, symbolic_expression=expr)
        else:
            try:
                if self.symbol is not None:
                    expr = self.symbol / other
                    descr = self.description + " / " + str(other)
                elif self.symbolic_expression is not None:
                    expr = (self.symbolic_expression) / other
                    descr = "(" + self.description + ") / " + str(other)
                else:
                    expr = None
                    descr = self.description + " / " + str(other)
                return ScalingParameter(res, description=descr, units=self.units, symbol=None, symbolic_expression=expr)
            except:
                return res

    def __rtruediv__(self, other):
        res = other / float(self)
        try:
            if self.symbol is not None:
                expr = other / self.symbol
                descr = str(other) + " / " + self.description
            elif self.symbolic_expression is not None:
                expr = other / (self.symbolic_expression)
                descr = str(other) + " / (" + self.description + ")"
            else:
                expr = None
                descr = str(other) + " / " + self.description
            return ScalingParameter(res, description=descr, units=self.units, symbol=None, symbolic_expression=expr)
        except:
            return res

    def __pow__(self, power, modulo=None):

        if modulo is not None:
            raise NotImplemented('ScalingParameter class: Modular exponentiation not implemented')

        res = float(self) ** power
        if int(power) == power:

            ul = self.units.split('][')
            ul[0] = ul[0][1:]
            ul[-1] = ul[-1][:-1]

            usl = list()
            for us in ul:
                up = us.split('^')
                if len(up) == 1:
                    up.append("1")

                usl.append(tuple(up))

            units_elements = list()
            for us in usl:
                units_elements.append(list((us[0], str(int(us[1]) * power))))

            units = list()
            for us in units_elements:
                if us is not None:
                    if int(us[1]) != 1:
                        units.append("[" + us[0] + "^" + us[1] + "]")
                    else:
                        units.append("[" + us[0] + "]")
            units = "".join(units)

            if self.symbolic_expression is not None:
                expr = (self.symbolic_expression) ** power
                descr = "(" + self.description + ") to the power "+str(power)
            elif self.symbol is not None:
                expr = self.symbol ** power
                descr = self.description + " to the power "+str(power)
            else:
                expr = None
                descr = self.description + " to the power "+str(power)

        else:
            power_fraction = Fraction(power)
            ul = self.units.split('][')
            ul[0] = ul[0][1:]
            ul[-1] = ul[-1][:-1]

            usl = list()
            for us in ul:
                up = us.split('^')
                if len(up) == 1:
                    up.append("1")

                usl.append(tuple(up))

            units_elements = list()
            for us in usl:
                new_power = int(us[1]) * power_fraction.numerator / power_fraction.denominator
                if int(new_power) == new_power:
                    units_elements.append(list((us[0], str(new_power))))
                else:
                    raise ArithmeticError("ScalingParameter class: Only support integer exponent in units")

            units = list()
            for us in units_elements:
                if us is not None:
                    if int(us[1]) != 1:
                        units.append("[" + us[0] + "^" + us[1] + "]")
                    else:
                        units.append("[" + us[0] + "]")
            units = "".join(units)
            if self.symbolic_expression is not None:
                expr = (self.symbolic_expression) ** power
                descr = "(" + self.description + ") to the power " + str(power)
            elif self.symbol is not None:
                expr = self.symbol ** power
                descr = self.description + " to the power " + str(power)
            else:
                expr = None
                descr = self.description + " to the power " + str(power)

        return ScalingParameter(res, description=descr, units=units, symbol=None, symbolic_expression=expr)


class Parameter(float):
    """Base class of model's parameter.

    Parameters
    ----------
    value: float
        Value of the parameter.
    input_dimensional: bool, optional
        Specify whether the value provided is dimensional or not. Default to `True`.
    units: str, optional
        The units of the provided value. Used to compute the conversion between dimensional and nondimensional
        value. Should be specified by joining atoms like `'[unit^power]'`, e.g '`[m^2][s^-2][Pa^-2]'`.
        Empty by default.
    scale_object: ScaleParams, optional
        A scale parameters object to compute the conversion between dimensional and nondimensional value.
        `None` by default. If `None`, cannot transform between dimensional and nondimentional value.
    description: str, optional
        String describing the parameter.
    symbol: ~sympy.core.symbol.Symbol, optional
        A `Sympy`_ symbol to represent the parameter in symbolic expressions.
    symbolic_expression: ~sympy.core.expr.Expr, optional
        A `Sympy`_ expression to represent a relationship to other parameters.
    return_dimensional: bool, optional
        Defined if the value returned by the parameter is dimensional or not. Default to `False`.

    Notes
    -----
    Parameter is immutable. Once instantiated, it cannot be altered. To create a new parameter, one must
    re-instantiate it.

    Warnings
    --------
    If no scale_object argument is provided, cannot transform between the dimensional and nondimentional value !

    .. _Sympy: https://www.sympy.org/
    """

    def __new__(cls, value, input_dimensional=True, units="", scale_object=None, description="",
                symbol=None, return_dimensional=False, symbolic_expression=None):

        no_scale = False

        if return_dimensional:
            if input_dimensional:
                evalue = value
            else:
                if scale_object is None:
                    return_dimensional = False
                    evalue = value
                    no_scale = True
                else:
                    evalue = value / cls._conversion_factor(units, scale_object)
        else:
            if input_dimensional:
                if scale_object is None:
                    return_dimensional = True
                    evalue = value
                    no_scale = True
                else:
                    try:
                        evalue = value * cls._conversion_factor(units, scale_object)
                    except:
                        print(description)
                        print(symbol)
                        print(units)
                        print(cls._conversion_factor(units, scale_object))
                        print(scale_object)
            else:
                evalue = value

        if no_scale:
            warnings.warn("Parameter configured to perform dimensional conversion " +
                          "but without specifying a ScaleParams object: Conversion disabled!")

        f = float.__new__(cls, evalue)
        f._input_dimensional = input_dimensional
        f._return_dimensional = return_dimensional
        f._units = units
        f._scale_object = scale_object
        f._description = description
        f._symbol = symbol
        f._symbolic_expression = symbolic_expression

        return f

    @property
    def dimensional_value(self):
        """float: Returns the dimensional value."""
        if self._return_dimensional:
            return self
        else:
            return self / self._nondimensionalization

    @property
    def nondimensional_value(self):
        """float: Returns the nondimensional value."""
        if self._return_dimensional:
            return self * self._nondimensionalization
        else:
            return self

    @property
    def symbol(self):
        """~sympy.core.symbol.Symbol: Returns the symbol of the parameter."""
        return self._symbol

    @property
    def symbolic_expression(self):
        """~sympy.core.expr.Expr: Returns the symbolic expression of the parameter."""
        if self._symbolic_expression is None and self._symbol is not None:
            return self._symbol
        else:
            return self._symbolic_expression
    
    @property
    def input_dimensional(self):
        """bool: Indicate if the provided value is dimensional or not."""
        return self._input_dimensional

    @property
    def return_dimensional(self):
        """bool: Indicate if the returned value is dimensional or not."""
        return self._return_dimensional

    @classmethod
    def _conversion_factor(cls, units, scale_object):
        factor = 1.

        ul = units.split('][')
        ul[0] = ul[0][1:]
        ul[-1] = ul[-1][:-1]

        for us in ul:
            up = us.split('^')
            if len(up) == 1:
                up.append("1")

            if up[0] == 'm':
                factor *= scale_object.L ** (-int(up[1]))
            elif up[0] == 's':
                factor *= scale_object.f0 ** (int(up[1]))
            elif up[0] == 'Pa':
                factor *= scale_object.deltap ** (-int(up[1]))

        return factor

    @property
    def units(self):
        """str: The units of the dimensional value."""
        return self._units

    @property
    def description(self):
        """str: Description of the parameter."""
        return self._description

    @property
    def _nondimensionalization(self):
        if self._scale_object is None:
            return 1.
        else:
            return self._conversion_factor(self._units, self._scale_object)

    def __add__(self, other):

        res = float(self) + other
        if isinstance(other, (Parameter, ScalingParameter)):
            if isinstance(other, Parameter) and self.return_dimensional != other.return_dimensional:
                raise ArithmeticError("Parameter class: Impossible to subtract a dimensional parameter with a non-dimensional one.")
            if self.units != other.units:
                raise ArithmeticError("Parameter class: Impossible to add two parameters with different units.")

            if self.symbolic_expression is None:
                if other.symbolic_expression is None:
                    if self.symbol is not None and other.symbol is not None:
                        expr = self.symbol + other.symbol
                    else:
                        expr = None
                    descr = self.description + " + " + other.description
                else:
                    if self.symbol is not None:
                        expr = self.symbol + (other.symbolic_expression)
                        descr = self.description + " + (" + other.description + ")"
                    else:
                        expr = None
                        descr = self.description + " + " + other.description
            else:
                if other.symbolic_expression is None:
                    if other.symbol is not None:
                        expr = (self.symbolic_expression) + other.symbol
                        descr = "(" + self.description + ") + " + other.description
                    else:
                        expr = None
                        descr = self.description + " + " + other.description
                else:
                    expr = (self.symbolic_expression) + (other.symbolic_expression)
                    descr = "(" + self.description + ") + (" + other.description + ")"

            return Parameter(res, input_dimensional=self.return_dimensional,
                             return_dimensional=self.return_dimensional, scale_object=self._scale_object,
                             description=descr, units=self.units, symbol=None, symbolic_expression=expr)
        else:
            try:
                if self.symbol is not None:
                    expr = self.symbol + other
                    descr = self.description + " + " + str(other)
                elif self.symbolic_expression is not None:
                    expr = (self.symbolic_expression) + other
                    descr = "(" + self.description + ") + " + str(other)
                else:
                    expr = None
                    descr = self.description + " + " + str(other)
                return Parameter(res, input_dimensional=self.return_dimensional,
                                 return_dimensional=self.return_dimensional, scale_object=self._scale_object,
                                 description=descr, units=self.units, symbol=None, symbolic_expression=expr)
            except:
                return res

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):

        res = float(self) - other
        if isinstance(other, (Parameter, ScalingParameter)):
            if isinstance(other, Parameter) and self.return_dimensional != other.return_dimensional:
                raise ArithmeticError("Parameter class: Impossible to subtract a dimensional parameter with a non-dimensional one.")
            if self.units != other.units:
                raise ArithmeticError("Parameter class: Impossible to subtract two parameters with different units.")
            if self.symbolic_expression is None:
                if other.symbolic_expression is None:
                    if self.symbol is not None and other.symbol is not None:
                        expr = self.symbol - other.symbol
                    else:
                        expr = None
                    descr = self.description + " - " + other.description
                else:
                    if self.symbol is not None:
                        expr = self.symbol - (other.symbolic_expression)
                        descr = self.description + " - (" + other.description + ")"
                    else:
                        expr = None
                        descr = self.description + " - " + other.description
            else:
                if other.symbolic_expression is None:
                    if other.symbol is not None:
                        expr = (self.symbolic_expression) - other.symbol
                        descr = "(" + self.description + ") - " + other.description
                    else:
                        expr = None
                        descr = self.description + " - " + other.description
                else:
                    expr = (self.symbolic_expression) - (other.symbolic_expression)
                    descr = "(" + self.description + ") - (" + other.description + ")"

            return Parameter(res, input_dimensional=self.return_dimensional,
                             return_dimensional=self.return_dimensional, scale_object=self._scale_object,
                             description=descr, units=self.units, symbol=None, symbolic_expression=expr)
        else:
            try:
                if self.symbol is not None:
                    expr = self.symbol - other
                    descr = self.description + " - " + str(other)
                elif self.symbolic_expression is not None:
                    expr = (self.symbolic_expression) - other
                    descr = "(" + self.description + ") - " + str(other)
                else:
                    expr = None
                    descr = self.description + " - " + str(other)
                return Parameter(res, input_dimensional=self.return_dimensional,
                                 return_dimensional=self.return_dimensional, scale_object=self._scale_object,
                                 description=descr, units=self.units, symbol=None, symbolic_expression=expr)
            except:
                return res

    def __rsub__(self, other):
        res = other - float(self)
        try:
            if self.symbol is not None:
                expr = other - self.symbol
                descr = str(other) + " - " + self.description
            elif self.symbolic_expression is not None:
                expr = other - (self.symbolic_expression)
                descr = str(other) + " - (" + self.description + ")"
            else:
                expr = None
                descr = str(other) + " - " + self.description
            return Parameter(res, input_dimensional=self.return_dimensional,
                             return_dimensional=self.return_dimensional, scale_object=self._scale_object,
                             description=descr, units=self.units, symbol=None, symbolic_expression=expr)
        except:
            return res

    def __mul__(self, other):

        res = float(self) * other
        if isinstance(other, (Parameter, ScalingParameter)):
            if hasattr(other, "units"):
                units = _combine_units(self.units, other.units, '+')
            else:
                units = ""

            if self.symbolic_expression is None:
                if other.symbolic_expression is None:
                    if self.symbol is not None and other.symbol is not None:
                        expr = self.symbol * other.symbol
                    else:
                        expr = None
                    descr = self.description + " * " + other.description
                else:
                    if self.symbol is not None:
                        expr = self.symbol * (other.symbolic_expression)
                        descr = self.description + " * (" + other.description + ")"
                    else:
                        expr = None
                        descr = self.description + " * " + other.description
            else:
                if other.symbolic_expression is None:
                    if other.symbol is not None:
                        expr = (self.symbolic_expression) * other.symbol
                        descr = "(" + self.description + ") * " + other.description
                    else:
                        expr = None
                        descr = self.description + " * " + other.description
                else:
                    expr = (self.symbolic_expression) * (other.symbolic_expression)
                    descr = "(" + self.description + ") * (" + other.description + ")"

            return Parameter(res, input_dimensional=self.return_dimensional, return_dimensional=self.return_dimensional,
                             scale_object=self._scale_object, description=descr, units=units, symbol=None,
                             symbolic_expression=expr)
        else:
            try:
                if self.symbol is not None:
                    expr = self.symbol * other
                    descr = self.description + " * " + str(other)
                elif self.symbolic_expression is not None:
                    expr = (self.symbolic_expression) * other
                    descr = "(" + self.description + ") * " + str(other)
                else:
                    expr = None
                    descr = self.description + " * " + str(other)
                return Parameter(res, input_dimensional=self.return_dimensional, return_dimensional=self.return_dimensional,
                                 scale_object=self._scale_object, description=descr, units=self.units, symbol=None,
                                 symbolic_expression=expr)
            except:
                return res

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):

        res = float(self) / other
        if isinstance(other, (ScalingParameter, Parameter)):
            units = _combine_units(self.units, other.units, '-')
            if self.symbolic_expression is None:
                if other.symbolic_expression is None:
                    if self.symbol is not None and other.symbol is not None:
                        expr = self.symbol / other.symbol
                    else:
                        expr = None
                    descr = self.description + " / " + other.description
                else:
                    if self.symbol is not None:
                        expr = self.symbol / (other.symbolic_expression)
                        descr = self.description + " / (" + other.description + ")"
                    else:
                        expr = None
                        descr = self.description + " / " + other.description
            else:
                if other.symbolic_expression is None:
                    if other.symbol is not None:
                        expr = (self.symbolic_expression) / other.symbol
                        descr = "(" + self.description + ") / " + other.description
                    else:
                        expr = None
                        descr = self.description + " / " + other.description
                else:
                    expr = (self.symbolic_expression) / (other.symbolic_expression)
                    descr = "(" + self.description + ") / (" + other.description + ")"

            return Parameter(res, input_dimensional=self.return_dimensional, return_dimensional=self.return_dimensional,
                             scale_object=self._scale_object, description=descr, units=units, symbol=None,
                             symbolic_expression=expr)
        else:
            try:
                if self.symbol is not None:
                    expr = self.symbol / other
                    descr = self.description + " / " + str(other)
                elif self.symbolic_expression is not None:
                    expr = (self.symbolic_expression) / other
                    descr = "(" + self.description + ") / " + str(other)
                else:
                    expr = None
                    descr = self.description + " / " + str(other)
                return Parameter(res, input_dimensional=self.return_dimensional, return_dimensional=self.return_dimensional,
                                 scale_object=self._scale_object, description=descr, units=self.units, symbol=None,
                                 symbolic_expression=expr)
            except:
                return res

    def __rtruediv__(self, other):
        res = other / float(self)
        try:
            if self.symbol is not None:
                expr = other / self.symbol
                descr = str(other) + " / " + self.description
            elif self.symbolic_expression is not None:
                expr = other / (self.symbolic_expression)
                descr = str(other) + " / (" + self.description + ")"
            else:
                expr = None
                descr = str(other) + " / " + self.description
            return Parameter(res, input_dimensional=self.return_dimensional,
                             return_dimensional=self.return_dimensional, scale_object=self._scale_object,
                             description=descr, units=self.units, symbol=None, symbolic_expression=expr)
        except:
            return res

    def __pow__(self, power, modulo=None):

        if modulo is not None:
            raise NotImplemented('Parameter class: Modular exponentiation not implemented')

        res = float(self) ** power
        if int(power) == power:

            ul = self.units.split('][')
            ul[0] = ul[0][1:]
            ul[-1] = ul[-1][:-1]

            usl = list()
            for us in ul:
                up = us.split('^')
                if len(up) == 1:
                    up.append("1")

                usl.append(tuple(up))

            units_elements = list()
            for us in usl:
                units_elements.append(list((us[0], str(int(us[1]) * power))))

            units = list()
            for us in units_elements:
                if us is not None:
                    if int(us[1]) != 1:
                        units.append("[" + us[0] + "^" + us[1] + "]")
                    else:
                        units.append("[" + us[0] + "]")
            units = "".join(units)

            if self.symbolic_expression is not None:
                expr = (self.symbolic_expression) ** power
                descr = "(" + self.description + ") to the power "+str(power)
            elif self.symbol is not None:
                expr = self.symbol ** power
                descr = self.description + " to the power "+str(power)
            else:
                expr = None
                descr = self.description + " to the power "+str(power)

        else:
            power_fraction = Fraction(power)
            ul = self.units.split('][')
            ul[0] = ul[0][1:]
            ul[-1] = ul[-1][:-1]

            usl = list()
            for us in ul:
                up = us.split('^')
                if len(up) == 1:
                    up.append("1")

                usl.append(tuple(up))

            units_elements = list()
            for us in usl:
                new_power = int(us[1]) * power_fraction.numerator / power_fraction.denominator
                if int(new_power) == new_power:
                    units_elements.append(list((us[0], str(int(new_power)))))
                else:
                    raise ArithmeticError("Parameter class: Only support integer exponent in units")

            units = list()
            for us in units_elements:
                if us is not None:
                    if int(us[1]) != 1:
                        units.append("[" + us[0] + "^" + us[1] + "]")
                    else:
                        units.append("[" + us[0] + "]")
            units = "".join(units)
            if self.symbolic_expression is not None:
                expr = (self.symbolic_expression) ** power
                descr = "(" + self.description + ") to the power "+str(power)
            elif self.symbol is not None:
                expr = self.symbol ** power
                descr = self.description + " to the power "+str(power)
            else:
                expr = None
                descr = self.description + " to the power "+str(power)

        return Parameter(res, input_dimensional=self.return_dimensional, return_dimensional=self.return_dimensional,
                         description=descr, units=units, scale_object=self._scale_object, symbol=None,
                         symbolic_expression=expr)


class ParametersArray(np.ndarray):
    """Base class of model's array of parameters.

    Parameters
    ----------
    values: list(float) or ~numpy.ndarray(float) or list(Parameter) or ~numpy.ndarray(Parameter) or list(ScalingParameter) or ~numpy.ndarray(ScalingParameter)
        Values of the parameter array.
    input_dimensional: bool, optional
        Specify whether the value provided is dimensional or not. Default to `True`.
    units: str, optional
        The units of the provided value. Used to compute the conversion between dimensional and nondimensional
        value. Should be specified by joining atoms like `'[unit^power]'`, e.g '`[m^2][s^-2][Pa^-2]'`.
        Empty by default.
    scale_object: ScaleParams, optional
        A scale parameters object to compute the conversion between dimensional and nondimensional value.
        `None` by default. If `None`, cannot transform between dimensional and nondimentional value.
    description: str or list(str) or array(str), optional
        String or an iterable of strings, describing the parameters.
        If an iterable, should have the same length or shape as `values`.
    symbols: ~sympy.core.symbol.Symbol or list(~sympy.core.symbol.Symbol) or ~numpy.ndarray(~sympy.core.symbol.Symbol), optional
        A `Sympy`_ symbol or an iterable of symbols, to represent the parameters in symbolic expressions.
        If an iterable, should have the same length or shape as `values`.
    symbolic_expressions: ~sympy.core.expr.Expr or list(~sympy.core.expr.Expr) or ~numpy.ndarray(~sympy.core.expr.Expr), optional
        A `Sympy`_ expression or an iterable of expressions, to represent a relationship to other parameters.
        If an iterable, should have the same length or shape as `values`.
    return_dimensional: bool, optional
        Defined if the value returned by the parameter is dimensional or not. Default to `False`.

    Warnings
    --------
    If no scale_object argument is provided, cannot transform between the dimensional and nondimensional value !

    .. _Sympy: https://www.sympy.org/
    """

    def __new__(cls, values, input_dimensional=True, units="", scale_object=None, description="",
                symbols=None, symbolic_expressions=None, return_dimensional=False):

        if isinstance(values, (tuple, list)):
            new_arr = np.empty(len(values), dtype=object)
            for i, val in enumerate(values):
                if isinstance(description, (tuple, list, np.ndarray)):
                    descr = description[i]
                else:
                    descr = description
                if isinstance(symbols, (tuple, list, np.ndarray)):
                    sy = symbols[i]
                else:
                    sy = symbols
                if isinstance(symbolic_expressions, (tuple, list, np.ndarray)):
                    expr = symbolic_expressions[i]
                else:
                    expr = symbolic_expressions
                new_arr[i] = Parameter(val, input_dimensional=input_dimensional, units=units, scale_object=scale_object, description=descr,
                                       return_dimensional=return_dimensional, symbol=sy, symbolic_expression=expr)
        else:
            if isinstance(values.flatten()[0], (Parameter, ScalingParameter)):
                new_arr = values.copy()
            else:
                new_arr = np.empty_like(values, dtype=object)
                for idx in np.ndindex(values.shape):
                    if isinstance(description, np.ndarray):
                        descr = description[idx]
                    else:
                        descr = description
                    if isinstance(symbols, np.ndarray):
                        sy = symbols[idx]
                    else:
                        sy = symbols
                    if isinstance(symbolic_expressions, np.ndarray):
                        expr = symbolic_expressions[idx]
                    else:
                        expr = symbolic_expressions
                    new_arr[idx] = Parameter(values[idx], input_dimensional=input_dimensional, units=units, scale_object=scale_object, description=descr,
                                             return_dimensional=return_dimensional, symbol=sy, symbolic_expression=expr)
        arr = np.asarray(new_arr).view(cls)
        arr._input_dimensional = input_dimensional
        arr._return_dimensional = return_dimensional
        arr._units = units
        arr._scale_object = scale_object

        return arr

    def __array_finalize__(self, arr):

        if arr is None:
            return

        self._input_dimensional = getattr(arr, '_input_dimensional', True)
        self._units = getattr(arr, '_units', "")
        self._return_dimensional = getattr(arr, '_return_dimensional', False)
        self._scale_object = getattr(arr, '_scale_object', None)

    @property
    def dimensional_values(self):
        """float: Returns the dimensional value."""
        if self._return_dimensional:
            return self
        else:
            return np.array(self / self._nondimensionalization)

    @property
    def nondimensional_values(self):
        """float: Returns the nondimensional value."""
        if self._return_dimensional:
            return np.array(self * self._nondimensionalization)
        else:
            return self

    @property
    def symbols(self):
        """~numpy.ndarray(~sympy.core.symbol.Symbol): Returns the symbol of the parameters in the array."""
        symbols = np.empty(self.shape, dtype=object)
        for idx in np.ndindex(self.shape):
            symbols[idx] = self[idx].symbol
        return symbols

    @property
    def symbolic_expressions(self):
        """~numpy.ndarray(~sympy.core.expr.Expr): Returns the symbolic expressions of the parameters in the array."""
        symbolic_expressions = np.empty(self.shape, dtype=object)
        for idx in np.ndindex(self.shape):
            symbolic_expressions[idx] = self[idx].symbolic_expression
        return symbolic_expressions

    @property
    def input_dimensional(self):
        """bool: Indicate if the provided value is dimensional or not."""
        return self._input_dimensional

    @property
    def return_dimensional(self):
        """bool: Indicate if the returned value is dimensional or not."""
        return self._return_dimensional

    @classmethod
    def _conversion_factor(cls, units, scale_object):
        factor = 1.

        ul = units.split('][')
        ul[0] = ul[0][1:]
        ul[-1] = ul[-1][:-1]

        for us in ul:
            up = us.split('^')
            if len(up) == 1:
                up.append("1")

            if up[0] == 'm':
                factor *= scale_object.L ** (-int(up[1]))
            elif up[0] == 's':
                factor *= scale_object.f0 ** (int(up[1]))
            elif up[0] == 'Pa':
                factor *= scale_object.deltap ** (-int(up[1]))

        return factor

    @property
    def units(self):
        """str: The units of the dimensional value."""
        return self._units

    @property
    def descriptions(self):
        """~numpy.ndarray(str): Description of the parameters in the array."""
        descr = np.empty(self.shape, dtype=object)
        for idx in np.ndindex(self.shape):
            descr[idx] = self[idx].description
        return descr

    @property
    def _nondimensionalization(self):
        if self._scale_object is None:
            return 1.
        else:
            return self._conversion_factor(self._units, self._scale_object)

    def __add__(self, other):
        if isinstance(other, (Parameter, ScalingParameter, float, int)):
            res = np.empty(self.shape, dtype=object)
            for idx in np.ndindex(self.shape):
                res[idx] = self[idx] + other
            item = res[idx]
            return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                   units=item.units, scale_object=self._scale_object)
        elif isinstance(other, ParametersArray):
            if other.shape == self.shape:  # Does not do broadcast
                res = np.empty(self.shape, dtype=object)
                for idx in np.ndindex(self.shape):
                    res[idx] = self[idx] + other[idx]
                item = res[idx]
                return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                       units=item.units, scale_object=self._scale_object)
            else:
                return self + other
        else:
            return self + other

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, (Parameter, ScalingParameter, float, int)):
            res = np.empty(self.shape, dtype=object)
            for idx in np.ndindex(self.shape):
                res[idx] = self[idx] - other
            item = res[idx]
            return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                   units=item.units, scale_object=self._scale_object)
        elif isinstance(other, ParametersArray):
            if other.shape == self.shape:  # Does not do broadcast
                res = np.empty(self.shape, dtype=object)
                for idx in np.ndindex(self.shape):
                    res[idx] = self[idx] - other[idx]
                item = res[idx]
                return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                       units=item.units, scale_object=self._scale_object)
            else:
                return self - other
        else:
            return self - other

    def __rsub__(self, other):
        if isinstance(other, (Parameter, ScalingParameter, float, int)):
            res = np.empty(self.shape, dtype=object)
            for idx in np.ndindex(self.shape):
                res[idx] = other - self[idx]
            item = res[idx]
            return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                   units=item.units, scale_object=self._scale_object)
        elif isinstance(other, ParametersArray):
            if other.shape == self.shape:  # Does not do broadcast
                res = np.empty(self.shape, dtype=object)
                for idx in np.ndindex(self.shape):
                    res[idx] = other - self[idx]
                item = res[idx]
                return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                       units=item.units, scale_object=self._scale_object)
            else:
                return other - self
        else:
            return other - self

    def __mul__(self, other):
        if isinstance(other, (Parameter, ScalingParameter, float, int)):
            res = np.empty(self.shape, dtype=object)
            for idx in np.ndindex(self.shape):
                res[idx] = self[idx] * other
            item = res[idx]
            return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                   units=item.units, scale_object=self._scale_object)
        elif isinstance(other, ParametersArray):
            if other.shape == self.shape:  # Does not do broadcast
                res = np.empty(self.shape, dtype=object)
                for idx in np.ndindex(self.shape):
                    res[idx] = self[idx] * other[idx]
                item = res[idx]
                return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                       units=item.units, scale_object=self._scale_object)
            else:
                return self * other
        else:
            return self * other

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, (Parameter, ScalingParameter, float, int)):
            res = np.empty(self.shape, dtype=object)
            for idx in np.ndindex(self.shape):
                res[idx] = self[idx] / other
            item = res[idx]
            return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                   units=item.units, scale_object=self._scale_object)
        elif isinstance(other, ParametersArray):
            if other.shape == self.shape:  # Does not do broadcast
                res = np.empty(self.shape, dtype=object)
                for idx in np.ndindex(self.shape):
                    res[idx] = self[idx] / other[idx]
                item = res[idx]
                return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                       units=item.units, scale_object=self._scale_object)
            else:
                return self / other
        else:
            return self / other

    def __rtruediv__(self, other):
        if isinstance(other, (Parameter, ScalingParameter, float, int)):
            res = np.empty(self.shape, dtype=object)
            for idx in np.ndindex(self.shape):
                res[idx] = other / self[idx]
            item = res[idx]
            return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                   units=item.units, scale_object=self._scale_object)
        elif isinstance(other, ParametersArray):
            if other.shape == self.shape:  # Does not do broadcast
                res = np.empty(self.shape, dtype=object)
                for idx in np.ndindex(self.shape):
                    res[idx] = other / self[idx]
                item = res[idx]
                return ParametersArray(res, input_dimensional=item.return_dimensional, return_dimensional=item.return_dimensional,
                                       units=item.units, scale_object=self._scale_object)
            else:
                return other / self
        else:
            return other / self


def _combine_units(units1, units2, operation):
    ul = units1.split('][')
    ul[0] = ul[0][1:]
    ul[-1] = ul[-1][:-1]
    ol = units2.split('][')
    ol[0] = ol[0][1:]
    ol[-1] = ol[-1][:-1]

    usl = list()
    for us in ul:
        up = us.split('^')
        if len(up) == 1:
            up.append("1")

        if up[0]:
            usl.append(tuple(up))

    osl = list()
    for os in ol:
        op = os.split('^')
        if len(op) == 1:
            op.append("1")

        if op[0]:
            osl.append(tuple(op))

    units_elements = list()
    for us in usl:
        new_us = [us[0]]
        i = 0
        for os in osl:
            if os[0] == us[0]:
                if operation == '-':
                    power = int(os[1]) - int(us[1])
                else:
                    power = int(os[1]) + int(us[1])
                del osl[i]
                break
            i += 1
        else:
            power = int(us[1])

        if power != 0:
            new_us.append(str(power))
            units_elements.append(new_us)

    if len(osl) != 0:
        units_elements += osl

    units = list()
    for us in units_elements:
        if us is not None:
            if int(us[1]) != 1:
                units.append("[" + us[0] + "^" + us[1] + "]")
            else:
                units.append("[" + us[0] + "]")
    return "".join(units)



"""
    Utility functions module
    ========================

    This module has some useful functions for the diagnostics.

"""

import numpy as np


def create_grid_basis(basis, X, Y, extra_subs=None):
    """Create an array from a basis with the basis functions evaluated on a grid.


    Parameters
    ----------
    basis: SymbolicBasis
        Symbolic basis to compute gridded representation from.
    X: ~numpy.ndarray
        The zonal x-coordinates of the grid points.
    Y: ~numpy.ndarray
        The meridional y-coordinates of the grid points.
    extra_subs: list(tuple), optional
        List of 2-tuples containing extra substitutions to be made with the functions providing
        the grib basis, before transforming them into python callable.
        The 2-tuples contain first a `Sympy`_  expression and then the value to substitute.

    Returns
    -------
    ~numpy.ndarray
        The gridded representation of the basis functions packed along the 0-th axis.

    """

    grid_basis = list()

    for func in basis.num_functions(extra_subs):
        grid = func(X, Y)
        if isinstance(grid, (int, float)):
            grid = np.ones_like(X) * grid
        grid_basis.append(grid)

    return np.array(grid_basis)



"""
    Inner products module (base class)
    ==================================

    Abstract base classes of the structure containing the inner products :math:`(\\cdot, \\, \\cdot)`  between the truncated set of basis functions :math:`\phi_i` for the ocean and
    land fields, and :math:`F_i` for the atmosphere fields (see :ref:`files/model/oro_model:Projecting the equations on a set of basis functions`).

    Description of the classes
    --------------------------

    The three classes computing and holding the inner products of the basis functions are:

    * :class:`AtmosphericInnerProducts`
    * :class:`OceanicInnerProducts`
    * :class:`GroundInnerProducts`

    Warnings
    --------

    These are `abstract base class`_, they must be subclassed and the functions to compute the inner products must be defined!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

"""

from abc import ABC, abstractmethod
import pickle


class AtmosphericInnerProducts(ABC):
    """Base class which contains all the atmospheric inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation.
    """

    def __init__(self):
        self._a = None
        self._u = None
        self._c = None
        self._b = None
        self._g = None
        self._d = None
        self._s = None
        self._z = None
        self._v = None
        self.stored = False

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the atmosphere  !
    # !-----------------------------------------------------!
    @property
    @abstractmethod
    def natm(self):
        """Number of atmospheric modes."""
        pass

    @abstractmethod
    def a(self, i, j):
        """Function to compute the matrix of the eigenvalues of the Laplacian (atmospheric): :math:`a_{i, j} = (F_i, {\\nabla}^2 F_j)`."""
        pass

    @abstractmethod
    def u(self, i, j):
        """Function to compute the matrix of inner product: :math:`u_{i, j} = (F_i, F_j)`."""
        pass

    @abstractmethod
    def b(self, i, j, k):
        """Function to compute the tensors holding the Jacobian inner products: :math:`b_{i, j, k} = (F_i, J(F_j, \\nabla^2 F_k))`."""
        pass

    @abstractmethod
    def c(self, i, j):
        """Function to compute the matrix of beta terms for the atmosphere: :math:`c_{i,j} = (F_i, \\partial_x F_j)`."""
        pass

    @abstractmethod
    def g(self, i, j, k):
        """Function to compute tensors holding the Jacobian inner products: :math:`g_{i,j,k} = (F_i, J(F_j, F_k))`."""
        pass

    @abstractmethod
    def s(self, i, j):
        """Function to compute the forcing (thermal) of the ocean on the atmosphere: :math:`s_{i,j} = (F_i, \\phi_j)`."""
        pass

    @abstractmethod
    def d(self, i, j):
        """Function to compute the forcing of the ocean on the atmosphere: :math:`d_{i,j} = (F_i, \\nabla^2 \\phi_j)`."""
        pass

    @abstractmethod
    def z(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing for the radiation lost by atmosphere to space & ground/ocean: :math:`z_{i,j,k,l,m} = (F_i, F_j F_k F_l F_m)`."""
        pass

    @abstractmethod
    def v(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing of the ocean on the atmosphere: :math:`v_{i,j,k,l,m} = (F_i, \\phi_j \\phi_k \\phi_l \\phi_m)`."""
        pass

    def save_to_file(self, filename, **kwargs):
        """Function to save the inner object to a file with the :mod:`pickle` module.

        Parameters
        ----------
        filename: str
            The file name where to save the inner product object.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'wb')
        pickle.dump(self.__dict__, f, **kwargs)
        f.close()

    def load_from_file(self, filename, **kwargs):
        """Function to load previously saved inner product object with the method :meth:`save_to_file`.

        Parameters
        ----------
        filename: str
            The file name where the inner product object was saved.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'rb')
        tmp_dict = pickle.load(f, **kwargs)
        f.close()

        self.__dict__.clear()
        self.__dict__.update(tmp_dict)


class OceanicInnerProducts(ABC):
    """Base class which contains all the oceanic inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation.
    """

    def __init__(self):
        self._M = None
        self._U = None
        self._N = None
        self._O = None
        self._C = None
        self._K = None
        self._W = None
        self._V = None
        self._Z = None

        self.stored = False

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the ocean       !
    # !-----------------------------------------------------!

    @property
    @abstractmethod
    def noc(self):
        """Number of oceanic modes."""
        pass

    @abstractmethod
    def M(self, i, j):
        """Function to compute the forcing of the ocean fields on the ocean: :math:`M_{i,j} = (\\phi_i, \\nabla^2 \\phi_j)`."""
        pass

    @abstractmethod
    def U(self, i, j):
        """Function to compute the inner products: :math:`U_{i,j} = (\\phi_i, \\phi_j)`."""
        pass

    @abstractmethod
    def N(self, i, j):
        """Function computing the beta term for the ocean: :math:`N_{i,j} = (\\phi_i, \\partial_x \\phi_j)`."""
        pass

    @abstractmethod
    def O(self, i, j, k):
        """Function to compute the temperature advection term (passive scalar): :math:`O_{i,j,k} = (\\phi_i, J(\\phi_j, \\phi_k))`"""
        pass

    @abstractmethod
    def C(self, i, j, k):
        """Function to compute the tensors holding the Jacobian inner products: :math:`C_{i,j,k} = (\\phi_i, J(\\phi_j,\\nabla^2 \\phi_k))`."""
        pass

    @abstractmethod
    def K(self, i, j):
        """Function to compute the forcing of the ocean by the atmosphere: :math:`K_{i,j} = (\\phi_i, \\nabla^2 F_j)`."""
        pass

    @abstractmethod
    def W(self, i, j):
        """Function to compute the short-wave radiative forcing of the ocean: :math:`W_{i,j} = (\\phi_i, F_j)`."""
        pass

    @abstractmethod
    def Z(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing from the atmosphere to the ocean: :math:`Z_{i,j,k,l,m} = (\\phi_i, F_j, F_k, F_l, F_m)`."""
        pass

    @abstractmethod
    def V(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing from the ocean to the atmosphere: :math:`V_{i,j,k,l,m} = (\\phi_i, \\phi_j, \\phi_k, \\phi_l, \\phi_m)`."""
        pass

    def save_to_file(self, filename, **kwargs):
        """Function to save the inner product object to a file with the :mod:`pickle` module.

        Parameters
        ----------
        filename: str
            The file name where to save the inner product object.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'wb')
        pickle.dump(self.__dict__, f, **kwargs)
        f.close()

    def load_from_file(self, filename, **kwargs):
        """Function to load previously saved inner product object with the method :meth:`save_to_file`.

        Parameters
        ----------
        filename: str
            The file name where the inner product object was saved.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'rb')
        tmp_dict = pickle.load(f, **kwargs)
        f.close()

        self.__dict__.clear()
        self.__dict__.update(tmp_dict)


class GroundInnerProducts(ABC):
    """Base class which contains all the ground inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation.
    """

    def __init__(self):
        self._M = None
        self._U = None
        self._N = None
        self._O = None
        self._C = None
        self._K = None
        self._W = None
        self._Z = None
        self._V = None

        self.stored = False

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the ocean       !
    # !-----------------------------------------------------!

    @property
    @abstractmethod
    def ngr(self):
        """Number of ground modes."""
        pass

    @abstractmethod
    def K(self, i, j):
        """Function to compute the forcing of the ocean by the atmosphere: :math:`K_{i,j} = (\\phi_i, \\nabla^2 F_j)`."""
        pass

    @abstractmethod
    def M(self, i, j):
        """Function to compute the forcing of the ocean fields on the ocean: :math:`M_{i,j} = (\\phi_i, \\nabla^2 \\phi_j)`."""
        pass

    @abstractmethod
    def U(self, i, j):
        """Function to compute the inner products: :math:`U_{i,j} = (\\phi_i, \\phi_j)`."""
        pass

    @abstractmethod
    def N(self, i, j):
        """Function computing the beta term for the ocean: :math:`N_{i,j} = (\\phi_i, \\partial_x \\phi_j)`."""
        pass

    @abstractmethod
    def O(self, i, j, k):
        """Function to compute the temperature advection term (passive scalar): :math:`O_{i,j,k} = (\\phi_i, J(\\phi_j, \\phi_k))`"""
        pass

    @abstractmethod
    def C(self, i, j, k):
        """Function to compute the tensors holding the Jacobian inner products: :math:`C_{i,j,k} = (\\phi_i, J(\\phi_j,\\nabla^2 \\phi_k))`."""
        pass

    @abstractmethod
    def W(self, i, j):
        """Function to compute the short-wave radiative forcing of the ocean: :math:`W_{i,j} = (\\phi_i, F_j)`."""
        pass

    @abstractmethod
    def V(self, i, j, k, l, m):
        pass

    @abstractmethod
    def Z(self, i, j, k, l, m):
        pass

    def save_to_file(self, filename, **kwargs):
        """Function to save the inner object to a file with the :mod:`pickle` module.

        Parameters
        ----------
        filename: str
            The file name where to save the inner product object.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'wb')
        pickle.dump(self.__dict__, f, **kwargs)
        f.close()

    def load_from_file(self, filename, **kwargs):
        """Function to load previously saved inner product object with the method :meth:`save_to_file`.

        Parameters
        ----------
        filename: str
            The file name where the inner product object was saved.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'rb')
        tmp_dict = pickle.load(f, **kwargs)
        f.close()

        self.__dict__.clear()
        self.__dict__.update(tmp_dict)



"""
    Diagnostic wind classes
    =======================

    Classes defining wind fields diagnostics.

    Description of the classes
    --------------------------

    * :class:`AtmosphericWindDiagnostic`: General base class for atmospheric wind diagnostic.
    * :class:`LowerLayerAtmosphericVWindDiagnostic`: Diagnostic giving the lower layer atmospheric V wind fields :math:`\\partial_x \\psi^3_{\\rm a}`.
    * :class:`LowerLayerAtmosphericUWindDiagnostic`: Diagnostic giving the lower layer atmospheric U wind fields :math:`- \\partial_y \\psi^3_{\\rm a}`.
    * :class:`MiddleAtmosphericVWindDiagnostic`: Diagnostic giving the middle atmospheric V wind fields :math:`\\partial_x \\psi_{\\rm a}`.
    * :class:`MiddleAtmosphericUWindDiagnostic`: Diagnostic giving the middle atmospheric U wind fields :math:`- \\partial_y \\psi_{\\rm a}`.
    * :class:`UpperLayerAtmosphericVWindDiagnostic`: Diagnostic giving the upper layer atmospheric V wind fields :math:`\\partial_x \\psi^1_{\\rm a}`.
    * :class:`UpperLayerAtmosphericUWindDiagnostic`: Diagnostic giving the upper layer atmospheric U wind fields :math:`- \\partial_y \\psi^1_{\\rm a}`.
    * :class:`LowerLayerAtmosphericWindIntensityDiagnostic`: Diagnostic giving the lower layer atmospheric horizontal wind intensity fields.
    * :class:`MiddleAtmosphericWindIntensityDiagnostic`: Diagnostic giving the middle atmospheric horizontal wind intensity fields.
    * :class:`UpperLayerAtmosphericWindIntensityDiagnostic`: Diagnostic giving the upper layer atmospheric horizontal wind intensity fields.
    * :class:`MiddleLayerVerticalVelocity`: Diagnostic giving the middle atmospheric layer vertical wind intensity fields.

"""

import warnings

import numpy as np
from numba import njit
import matplotlib.pyplot as plt

from qgs.diagnostics.differential import DifferentialFieldDiagnostic
from qgs.diagnostics.util import create_grid_basis
from qgs.functions.tendencies import create_tendencies, create_atmo_thermo_tendencies


class AtmosphericWindDiagnostic(DifferentialFieldDiagnostic):
    """General base class for atmospheric wind fields diagnostic.
    Provide a spatial gridded representation of the fields.
    This is an `abstract base class`_, it must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        if not hasattr(self, 'type'):
            self.type = None

        DifferentialFieldDiagnostic.__init__(self, model_params, dimensional)
        self._configure(delta_x=delta_x, delta_y=delta_y)

        self._plot_units = r" (in " + r'm s$^{-1}$' + r")"
        self._default_plot_kwargs['cmap'] = plt.get_cmap('hsv_r')
        self._color_bar_format = False

    def _compute_grid(self, delta_x=None, delta_y=None):

        if delta_x is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericWindDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            xwn = [ams[i][0] for i in range(len(ams))]
            mxwn = max(xwn)
            n_point_x = 4 * mxwn + 2
        else:
            n_point_x = int(np.ceil((2 * np.pi / self._model_params.scale_params.n) / delta_x) + 1)

        if delta_y is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericWindDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            ywn = [ams[i][1] for i in range(len(ams))]
            mywn = max(ywn)
            n_point_y = 4 * mywn + 2
        else:
            n_point_y = int(np.ceil(np.pi / delta_y) + 1)

        x = np.linspace(0., 2 * np.pi / self._model_params.scale_params.n, n_point_x)
        y = np.linspace(0., np.pi, n_point_y)
        self._X, self._Y = np.meshgrid(x, y)

    def _configure(self, delta_x=None, delta_y=None):

        basis = self._model_params.atmospheric_basis

        if self.type == "V":
            self._configure_differential_grid(basis, "dx", 1, delta_x, delta_y)

        elif self.type == "U":
            self._configure_differential_grid(basis, "dy", 1, delta_x, delta_y)

        elif self.type == "W":
            self._compute_grid(delta_x, delta_y)
            self._grid_basis = create_grid_basis(basis, self._X, self._Y, self._subs)

        elif self.type is None:
            warnings.warn("AtmosphericWindDiagnostic: Basis type note specified." +
                          " Unable to configure the diagnostic properly.")
            return 1

        else:
            self._grid_basis = None

        if self._orography and self._X is not None and self._Y is not None:
            if self._model_params.ground_params.orographic_basis == "atmospheric":
                self._oro_basis = create_grid_basis(basis, self._X, self._Y, self._subs)
            else:
                self._oro_basis = create_grid_basis(self._model_params.ground_basis, self._X, self._Y, self._subs)
        else:
            self._oro_basis = None


class LowerLayerAtmosphericVWindDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the lower layer atmospheric V wind fields :math:`\\partial_x \\psi^3_{\\rm a}`.
    Computed as :math:`\\partial_x \\psi^3_{\\rm a} = \\partial_x \\psi_{\\rm a} - \\partial_x \\theta_{\\rm a}` where :math:`\\psi_{\\rm a}` and :math:`\\theta_{\\rm a}` are respectively the barotropic and baroclinic streamfunctions.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = "V"

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric V wind in the lower layer'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        psi3 = psi - theta

        if dimensional:
            self._diagnostic_data = psi3 * self._model_params.streamfunction_scaling / self._model_params.scale_params.L
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi3
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class LowerLayerAtmosphericUWindDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the lower layer atmospheric U wind fields :math:`- \\partial_y \\psi^3_{\\rm a}`.
    Computed as :math:`- \\partial_y \\psi^3_{\\rm a} = - \\partial_y \\psi_{\\rm a} + \\partial_y \\theta_{\\rm a}` where :math:`\\psi_{\\rm a}` and :math:`\\theta_{\\rm a}` are respectively the barotropic and baroclinic streamfunctions.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = "U"

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric U wind in the lower layer'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        psi3 = psi - theta

        if dimensional:
            self._diagnostic_data = - psi3 * self._model_params.streamfunction_scaling / self._model_params.scale_params.L
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = - psi3
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class MiddleAtmosphericVWindDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the middle atmospheric V wind fields :math:`\\partial_x \\psi_{\\rm a}`
    where :math:`\\psi_{\\rm a}` is the barotropic streamfunction.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations
    and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = "V"

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric V wind in the middle of the atmosphere'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = psi * self._model_params.streamfunction_scaling / self._model_params.scale_params.L
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class MiddleAtmosphericUWindDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the middle atmospheric U wind fields :math:`- \\partial_y \\psi_{\\rm a}` where
    :math:`\\psi_{\\rm a}` is the barotropic streamfunction.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations
    and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = "U"

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric U wind in the middle of the atmosphere'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = - psi * self._model_params.streamfunction_scaling / self._model_params.scale_params.L
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = - psi
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class UpperLayerAtmosphericVWindDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the upper layer atmospheric V wind fields :math:`\\partial_x \\psi^1_{\\rm a}`.
    Computed as :math:`\\partial_x \\psi^1_{\\rm a} = \\partial_x \\psi_{\\rm a} + \\partial_x \\theta_{\\rm a}` where :math:`\\psi_{\\rm a}` and :math:`\\theta_{\\rm a}` are respectively the barotropic and baroclinic streamfunctions.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = "V"

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric V wind in the upper layer'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        psi1 = psi + theta

        if dimensional:
            self._diagnostic_data = psi1 * self._model_params.streamfunction_scaling / self._model_params.scale_params.L
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi1
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class UpperLayerAtmosphericUWindDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the upper layer atmospheric U wind fields :math:`- \\partial_y \\psi^1_{\\rm a}`.
    Computed as :math:`- \\partial_y \\psi^1_{\\rm a} = - \\partial_y \\psi_{\\rm a} - \\partial_y \\theta_{\\rm a}` where :math:`\\psi_{\\rm a}` and :math:`\\theta_{\\rm a}` are respectively the barotropic and baroclinic streamfunctions.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = "U"

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric U wind in the upper layer'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        psi1 = psi + theta

        if dimensional:
            self._diagnostic_data = - psi1 * self._model_params.streamfunction_scaling / self._model_params.scale_params.L
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = - psi1
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class LowerLayerAtmosphericWindIntensityDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the lower layer atmospheric horizontal wind intensity fields.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = ""

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric wind intensity in the lower layer'
        self._udiag = LowerLayerAtmosphericUWindDiagnostic(model_params, delta_x, delta_y, dimensional)
        self._vdiag = LowerLayerAtmosphericVWindDiagnostic(model_params, delta_x, delta_y, dimensional)

    def _get_diagnostic(self, dimensional):

        self._udiag.set_data(self._time, self._data)
        self._vdiag.set_data(self._time, self._data)

        U = self._udiag._get_diagnostic(dimensional)
        V = self._vdiag._get_diagnostic(dimensional)

        self._diagnostic_data = np.sqrt(U**2 + V**2)

        if dimensional:
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class MiddleAtmosphericWindIntensityDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the middle atmospheric horizontal wind intensity fields.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = ""

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Wind intensity in the middle of the atmosphere'
        self._udiag = MiddleAtmosphericUWindDiagnostic(model_params, delta_x, delta_y, dimensional)
        self._vdiag = MiddleAtmosphericVWindDiagnostic(model_params, delta_x, delta_y, dimensional)

    def _get_diagnostic(self, dimensional):

        self._udiag.set_data(self._time, self._data)
        self._vdiag.set_data(self._time, self._data)

        U = self._udiag._get_diagnostic(dimensional)
        V = self._vdiag._get_diagnostic(dimensional)

        self._diagnostic_data = np.sqrt(U**2 + V**2)

        if dimensional:
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class UpperLayerAtmosphericWindIntensityDiagnostic(AtmosphericWindDiagnostic):
    """Diagnostic giving the lower layer atmospheric horizontal wind intensity fields.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = ""

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric wind intensity in the upper layer'
        self._udiag = UpperLayerAtmosphericUWindDiagnostic(model_params, delta_x, delta_y, dimensional)
        self._vdiag = UpperLayerAtmosphericVWindDiagnostic(model_params, delta_x, delta_y, dimensional)

    def _get_diagnostic(self, dimensional):

        self._udiag.set_data(self._time, self._data)
        self._vdiag.set_data(self._time, self._data)

        U = self._udiag._get_diagnostic(dimensional)
        V = self._vdiag._get_diagnostic(dimensional)

        U = self._udiag._get_diagnostic(dimensional)
        V = self._vdiag._get_diagnostic(dimensional)

        self._diagnostic_data = np.sqrt(U**2 + V**2)

        if dimensional:
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class MiddleLayerVerticalVelocity(AtmosphericWindDiagnostic):
    """Diagnostic giving the middle atmospheric layer vertical wind intensity fields :math:`\\omega`.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        self.type = "W"

        AtmosphericWindDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric vertical wind in the middle layer'
        self._plot_units = r" (in " + r'Pa s$^{-1}$' + r")"

        self._f, _ = create_tendencies(model_params)
        self._f_thermo = create_atmo_thermo_tendencies(model_params)

    def set_data(self, time, data):

        self._time = time
        self._data = _compute_omega_term(time, data, self._f, self._f_thermo)
        self._data = self._data / self._model_params.atmospheric_params.sig0

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        omega = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis, 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = omega * self._model_params.scale_params.deltap * self._model_params.scale_params.f0
        else:
            self._diagnostic_data = omega

        return self._diagnostic_data


@njit
def _compute_omega_term(time, data, func, thermo_func):
    tendencies = np.zeros_like(data)
    thermo_tendencies = np.zeros_like(tendencies)
    for i in range(data.shape[-1]):
        tendencies[:, i] = func(time[i], data[:, i])
        thermo_tendencies[:, i] = thermo_func(time[i], data[:, i])

    return tendencies - thermo_tendencies


if __name__ == '__main__':
    from qgs.params.params import QgParams
    from qgs.integrators.integrator import RungeKuttaIntegrator

    pars = QgParams()
    pars.set_atmospheric_channel_fourier_modes(2, 2)
    f, Df = create_tendencies(pars)
    integrator = RungeKuttaIntegrator()
    integrator.set_func(f)
    ic = np.random.rand(pars.ndim) * 0.1
    integrator.integrate(0., 10000., 0.1, ic=ic, write_steps=5)
    time, traj = integrator.get_trajectories()
    integrator.terminate()

    dx_psi3 = LowerLayerAtmosphericVWindDiagnostic(pars)
    dx_psi3(time, traj)

    dy_psi3 = LowerLayerAtmosphericUWindDiagnostic(pars)
    dy_psi3(time, traj)

    dx_psi = MiddleAtmosphericVWindDiagnostic(pars)
    dx_psi(time, traj)

    dy_psi = MiddleAtmosphericUWindDiagnostic(pars)
    dy_psi(time, traj)

    dx_psi1 = UpperLayerAtmosphericVWindDiagnostic(pars)
    dx_psi1(time, traj)

    dy_psi1 = UpperLayerAtmosphericUWindDiagnostic(pars)
    dy_psi1(time, traj)

    psi3_wind = LowerLayerAtmosphericWindIntensityDiagnostic(pars)
    psi3_wind(time, traj)

    psi_wind = MiddleAtmosphericWindIntensityDiagnostic(pars)
    psi_wind(time, traj)

    psi1_wind = UpperLayerAtmosphericWindIntensityDiagnostic(pars)
    psi1_wind(time, traj)

    vert_wind = MiddleLayerVerticalVelocity(pars)
    vert_wind(time, traj)



"""
    Diagnostic vorticity classes
    ============================

    Classes defining vorticity fields diagnostics.

    Description of the classes
    --------------------------

    * :class:`AtmosphericVorticityDiagnostic`: General base class for atmospheric vorticity diagnostic.
    * :class:`LowerLayerAtmosphericVorticityDiagnostic`: Diagnostic giving the lower layer atmospheric vorticity fields :math:`\\nabla^2 \\psi^3_{\\rm a}`.
    * :class:`MiddleAtmosphericVorticityDiagnostic`: Diagnostic giving the middle atmospheric vorticity fields :math:`\\nabla^2 \\psi_{\\rm a}`.
    * :class:`UpperLayerAtmosphericVorticityDiagnostic`: Diagnostic giving the upper layer atmospheric vorticity fields :math:`\\nabla^2 \\psi^1_{\\rm a}`.
    * :class:`LowerLayerAtmosphericPotentialVorticityDiagnostic`: Diagnostic giving the lower layer atmospheric potential vorticity fields :math:`\\nabla^2 \\psi^3_{\\rm a} + f_0 + \\beta\\, y + \\frac{f_0^2}{\\sigma_0\\, \\delta p^2} \\theta_{\\rm a}`.
    * :class:`UpperLayerAtmosphericPotentialVorticityDiagnostic`: Diagnostic giving the upper layer atmospheric potential vorticity fields :math:`\\nabla^2 \\psi^1_{\\rm a} + f_0 + \\beta\\, y - \\frac{f_0^2}{\\sigma_0\\, \\delta p^2} \\theta_{\\rm a}`.

"""

import warnings

from qgs.diagnostics.differential import LaplacianFieldDiagnostic
from qgs.diagnostics.util import create_grid_basis

import numpy as np
import matplotlib.pyplot as plt


class AtmosphericVorticityDiagnostic(LaplacianFieldDiagnostic):
    """General base class for atmospheric vorticity fields diagnostic.
    Provide a spatial gridded representation of the fields.
    This is an `abstract base class`_, it must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        LaplacianFieldDiagnostic.__init__(self, model_params, dimensional)
        self._configure(delta_x=delta_x, delta_y=delta_y)

        self._plot_units = r" (in " + r's$^{-1}$' + r")"
        self._default_plot_kwargs['cmap'] = plt.get_cmap('hsv_r')
        self._color_bar_format = False

    def _compute_grid(self, delta_x=None, delta_y=None):

        if delta_x is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericVorticityDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            xwn = [ams[i][0] for i in range(len(ams))]
            mxwn = max(xwn)
            n_point_x = 4 * mxwn + 2
        else:
            n_point_x = int(np.ceil((2 * np.pi / self._model_params.scale_params.n) / delta_x) + 1)

        if delta_y is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericVorticityDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            ywn = [ams[i][1] for i in range(len(ams))]
            mywn = max(ywn)
            n_point_y = 4 * mywn + 2
        else:
            n_point_y = int(np.ceil(np.pi / delta_y) + 1)

        x = np.linspace(0., 2 * np.pi / self._model_params.scale_params.n, n_point_x)
        y = np.linspace(0., np.pi, n_point_y)
        self._X, self._Y = np.meshgrid(x, y)

    def _configure(self, delta_x=None, delta_y=None):

        basis = self._model_params.atmospheric_basis
        self._configure_laplacian_grid(basis, delta_x, delta_y)

        if self._orography and self._X is not None and self._Y is not None:
            if self._model_params.ground_params.orographic_basis == "atmospheric":
                self._oro_basis = create_grid_basis(basis, self._X, self._Y, self._subs)
            else:
                self._oro_basis = create_grid_basis(self._model_params.ground_basis, self._X, self._Y, self._subs)
        else:
            self._oro_basis = None


class LowerLayerAtmosphericVorticityDiagnostic(AtmosphericVorticityDiagnostic):
    """Diagnostic giving the lower layer atmospheric vorticity fields :math:`\\nabla^2 \\psi^3_{\\rm a}`.
    Computed as :math:`\\nabla^2 \\psi^3_{\\rm a} = \\nabla^2 \\psi_{\\rm a} - \\nabla^2 \\theta_{\\rm a}` where :math:`\\psi_{\\rm a}` and :math:`\\theta_{\\rm a}` are respectively the barotropic and baroclinic streamfunctions.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericVorticityDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric vorticity in the lower layer'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        psi3 = psi - theta

        if dimensional:
            self._diagnostic_data = psi3 * self._model_params.streamfunction_scaling / (self._model_params.scale_params.L ** 2)
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi3
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class MiddleAtmosphericVorticityDiagnostic(AtmosphericVorticityDiagnostic):
    """Diagnostic giving the middle atmospheric vorticity fields :math:`\\nabla^2 \\psi_{\\rm a}`
    where :math:`\\psi_{\\rm a}` is the barotropic streamfunction.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations
    and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericVorticityDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric vorticity in the middle of the atmosphere'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = psi * self._model_params.streamfunction_scaling / (self._model_params.scale_params.L ** 2)
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class UpperLayerAtmosphericVorticityDiagnostic(AtmosphericVorticityDiagnostic):
    """Diagnostic giving the upper layer atmospheric vorticity fields :math:`\\nabla^2 \\psi^1_{\\rm a}`.
    Computed as :math:`\\nabla^2 \\psi^1_{\\rm a} = \\nabla^2 \\psi_{\\rm a} + \\nabla^2 \\theta_{\\rm a}` where :math:`\\psi_{\\rm a}` and :math:`\\theta_{\\rm a}` are respectively the barotropic and baroclinic streamfunctions.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericVorticityDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric vorticity in the upper layer'

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        psi1 = psi + theta

        if dimensional:
            self._diagnostic_data = psi1 * self._model_params.streamfunction_scaling / (self._model_params.scale_params.L ** 2)
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi1
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class UpperLayerAtmosphericPotentialVorticityDiagnostic(AtmosphericVorticityDiagnostic):
    """Diagnostic giving the upper layer atmospheric potential vorticity fields :math:`\\nabla^2 \\psi^1_{\\rm a} + f_0 + \\beta\\, y - \\frac{f_0^2}{\\sigma_0\\, \\delta p^2} \\theta_{\\rm a}`.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericVorticityDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric potential vorticity in the upper layer'

        self._vorticity = UpperLayerAtmosphericVorticityDiagnostic(model_params, delta_x, delta_y, dimensional)

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        self._vorticity.set_data(self._time, self._data)
        vorticity = self._vorticity._get_diagnostic(dimensional)

        if dimensional:
            self._diagnostic_data = vorticity
            self._diagnostic_data += self._model_params.scale_params.f0 + self._model_params.scale_params.beta.dimensional_value * self._Y * self._model_params.scale_params.L
            self._diagnostic_data -= (self._model_params.scale_params.f0 ** 2) * (theta * self._model_params.streamfunction_scaling) \
                                     / (self._model_params.atmospheric_params.sig0.dimensional_value * self._model_params.scale_params.deltap ** 2)
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = vorticity + 1 + self._model_params.scale_params.beta * self._Y - theta / self._model_params.atmospheric_params.sig0
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class LowerLayerAtmosphericPotentialVorticityDiagnostic(AtmosphericVorticityDiagnostic):
    """Diagnostic giving the lower layer atmospheric potential vorticity fields :math:`\\nabla^2 \\psi^3_{\\rm a} + f_0 + \\beta\\, y + \\frac{f_0^2}{\\sigma_0\\, \\delta p^2} \\theta_{\\rm a}`.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericVorticityDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric potential vorticity in the lower layer'

        self._vorticity = LowerLayerAtmosphericVorticityDiagnostic(model_params, delta_x, delta_y, dimensional)

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        self._vorticity.set_data(self._time, self._data)
        vorticity = self._vorticity._get_diagnostic(dimensional)

        if dimensional:
            self._diagnostic_data = vorticity
            self._diagnostic_data += self._model_params.scale_params.f0 + self._model_params.scale_params.beta.dimensional_value * self._Y * self._model_params.scale_params.L
            self._diagnostic_data += (self._model_params.scale_params.f0 ** 2) * (theta * self._model_params.streamfunction_scaling) \
                                     / (self._model_params.atmospheric_params.sig0.dimensional_value * self._model_params.scale_params.deltap ** 2)
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = vorticity + 1 + self._model_params.scale_params.beta * self._Y + theta / self._model_params.atmospheric_params.sig0
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data



"""
    Fourier Basis definition module
    ===============================

    Classes and functions defining Fourier basis of functions (`Fourier modes`_) and used to configure the model.
    (see :ref:`files/model/oro_model:Projecting the equations on a set of basis functions`).

    Description of the classes
    --------------------------

    * :class:`ChannelFourierBasis`: Fourier basis defined on a zonally perdiodic channel, with no-flux boundary conditions in the meridional direction :math:`y`.
    * :class:`BasinFourierBasis`: Fourier basis defined on a closed basin, with no-flux boundary conditions in both the zonal and meridional direction :math:`x` and :math:`y`.

    .. _Fourier modes: https://en.wikipedia.org/wiki/Fourier_series

"""
import numpy as np

from qgs.basis.base import SymbolicBasis

from sympy import symbols, sin, cos, sqrt

_x, _y = symbols('x y')
_n = symbols('n', positive=True)


class ChannelFourierBasis(SymbolicBasis):
    """Fourier basis defined on a zonally perdiodic channel, with no-flux boundary conditions in the meridional
    direction :math:`y`.

    Parameters
    ----------
    spectral_blocks: ~numpy.ndarray(int)
        Spectral blocks detailing the modes :math:`x`- and :math:`y`-wavenumber.
        Array of shape (nblocks, 2), where `nblocks` is the number of spectral blocks.
    aspect_ratio: float
        Spatial domain aspect ratio, :math:`n = 2 L_y/L_x` .
    """

    def __init__(self, spectral_blocks, aspect_ratio):

        SymbolicBasis.__init__(self)
        self.substitutions.append((_n, aspect_ratio))

        awavenum = channel_wavenumbers(spectral_blocks)

        for wv in awavenum:

            mode_eq = fourier_functions(wv)

            if mode_eq is not None:
                self.functions.append(mode_eq)


class BasinFourierBasis(SymbolicBasis):
    """Fourier basis defined on a closed basin, with no-flux boundary conditions in both the zonal and meridional
    direction :math:`x` and :math:`y`.

    Parameters
    ----------
    spectral_blocks: ~numpy.ndarray(int)
        Spectral blocks detailing the modes :math:`x`- and :math:`y`-wavenumber.
        Array of shape (nblocks, 2), where `nblocks` is the number of spectral blocks.
    aspect_ratio: float
        Spatial domain aspect ratio, :math:`n = 2 L_y/L_x` .
    """

    def __init__(self, spectral_blocks, aspect_ratio):

        SymbolicBasis.__init__(self)
        self.substitutions.append((_n, aspect_ratio))

        owavenum = basin_wavenumbers(spectral_blocks)

        for wv in owavenum:

            mode_eq = fourier_functions(wv)

            if mode_eq is not None:
                self.functions.append(mode_eq)


def contiguous_basin_basis(nxmax, nymax, aspect_ratio):
    """Function that returns the basis for contiguous spectral blocks of modes on a closed basin.

    Parameters
    ----------
    nxmax: int
        Maximum x-wavenumber to fill the spectral block up to.
    nymax: int
        Maximum :math:`y`-wavenumber to fill the spectral block up to.
    aspect_ratio: float
        Spatial domain aspect ratio, :math:`n = 2 L_y/L_x` .

    Returns
    -------
    BasinFourierBasis
        The closed basin contiguous basis up to the specified spectral truncation.
    """

    spectral_blocks = np.zeros((nxmax * nymax, 2), dtype=int)
    i = 0
    for nx in range(1, nxmax + 1):
        for ny in range(1, nymax+1):
            spectral_blocks[i, 0] = nx
            spectral_blocks[i, 1] = ny
            i += 1

    return BasinFourierBasis(spectral_blocks, aspect_ratio)


def contiguous_channel_basis(nxmax, nymax, aspect_ratio):
    """Function that returns the basis for contiguous spectral blocks of modes on a channel.

    Parameters
    ----------
    nxmax: int
        Maximum x-wavenumber to fill the spectral block up to.
    nymax: int
        Maximum :math:`y`-wavenumber to fill the spectral block up to.
    aspect_ratio: float
        Spatial domain aspect ratio, :math:`n = 2 L_y/L_x` .

    Returns
    -------
    ChannelFourierBasis
        The channel contiguous basis up to the specified spectral truncation.
    """

    spectral_blocks = np.zeros((nxmax * nymax, 2), dtype=int)
    i = 0
    for nx in range(1, nxmax + 1):
        for ny in range(1, nymax+1):
            spectral_blocks[i, 0] = nx
            spectral_blocks[i, 1] = ny
            i += 1

    return ChannelFourierBasis(spectral_blocks, aspect_ratio)


def fourier_functions(wave_number):
    """Function that return Fourier modes expressions:

    * `'A'` for a function of the form :math:`F^A_{P} (x, y) =  \sqrt{2}\, \cos(P y) = \sqrt{2}\, \cos(n_y\, y)`
    * `'K'` for a function of the form :math:`F^K_{M,P} (x, y) =  2\cos(M nx)\, \sin(P y) = 2\cos(n_x\,  n\, x)\, \sin(n_y\, y)`
    * `'L'` for a function of the form :math:`F^L_{H,P} (x, y) = 2\sin(H nx)\, \sin(P y) = 2\sin(n_x\, n \,x)\, \sin(n_y\, y)`

    Parameters
    ----------
    wave_number: WaveNumber
        The wavenumber and type information of the mode to be returned.

    Returns
    -------
    `Sympy`_ expression
        Symbolic expression of the mode.

    .. _Sympy: https://www.sympy.org/

    """
    mode_eq = None
    if wave_number.type == 'A':
        mode_eq = sqrt(2) * cos(wave_number.ny * _y)
    elif wave_number.type == 'K':
        mode_eq = 2 * cos(wave_number.nx * _n * _x) * sin(wave_number.ny * _y)
    elif wave_number.type == 'L':
        mode_eq = 2 * sin(wave_number.nx * _n * _x) * sin(wave_number.ny * _y)

    return mode_eq


class WaveNumber(object):
    """Class to define model base functions wavenumber. The basis function available are:

    * `'A'` for a function of the form :math:`F^A_{P} (x, y) =  \sqrt{2}\, \cos(P y) = \sqrt{2}\, \cos(n_y\, y)`
    * `'K'` for a function of the form :math:`F^K_{M,P} (x, y) =  2\cos(M nx)\, \sin(P y) = 2\cos(n_x\,  n\, x)\, \sin(n_y\, y)`
    * `'L'` for a function of the form :math:`F^L_{H,P} (x, y) = 2\sin(H nx)\, \sin(P y) = 2\sin(n_x\, n \,x)\, \sin(n_y\, y)`

    where :math:`x` and :math:`y` are the nondimensional model's domain coordinates (see :ref:`files/model/oro_model:Projecting the equations on a set of basis functions`).

    Parameters
    ----------
    function_type: str
        One character string to define the type of basis function. It can be `'A'`, `'K'` or `'L'`.
    P: int
        The :math:`y` wavenumber integer.
    M: int
        The :math:`x` wavenumber integer.
    H: int
        The :math:`x` wavenumber integer.
    nx: float
        The :math:`x` wavenumber.
    ny: float
        The :math:`y` wavenumber.

    Attributes
    ----------
    type: str
        One character string to define the type of basis function. It can be `'A'`, `'K'` or `'L'`.
    P: int
        The :math:`y` wavenumber integer.
    M: int
        The :math:`x` wavenumber integer.
    H: int
        The :math:`x` wavenumber integer.
    nx: float
        The :math:`x` wavenumber.
    ny: float
        The :math:`y` wavenumber.

    """

    def __init__(self, function_type, P, M, H, nx, ny):
        self.type = function_type
        self.P = P
        self.M = M
        self.H = H
        self.nx = nx
        self.ny = ny

    def __repr__(self):
        return "type = {}, P = {}, M= {},H={}, nx= {}, ny={}".format(self.type, self.P, self.M, self.H, self.nx, self.ny)


def channel_wavenumbers(spectral_blocks):
    """Functions that returns the :class:`WaveNumber` objects corresponding to a given list of spectral blocks for a
    channel-like spatial domain.

    Parameters
    ----------
    spectral_blocks: ~numpy.ndarray(int)
        Spectral blocks detailing the modes :math:`x`- and :math:`y`-wavenumber.
        Array of shape (nblocks, 2), where `nblocks` is the number of spectral blocks.

    Returns
    -------
    ~numpy.ndarray(WaveNumber)
        The array of wavenumber objects corresponding to the given spectral blocks.
    """
    # initialization of the variables
    wave_numbers = list()

    # Atmospheric wavenumbers definition

    for i in range(spectral_blocks.shape[0]):  # function type is limited to AKL for the moment: atmosphere is a channel

        if spectral_blocks[i, 0] == 1:
            wave_numbers.append(WaveNumber('A', spectral_blocks[i, 1], 0, 0, 0, spectral_blocks[i, 1]))
            wave_numbers.append(WaveNumber('K', spectral_blocks[i, 1], spectral_blocks[i, 0], 0, spectral_blocks[i, 0], spectral_blocks[i, 1]))
            wave_numbers.append(WaveNumber('L', spectral_blocks[i, 1], 0, spectral_blocks[i, 0], spectral_blocks[i, 0], spectral_blocks[i, 1]))
        else:
            wave_numbers.append(WaveNumber('K', spectral_blocks[i, 1], spectral_blocks[i, 0], 0, spectral_blocks[i, 0], spectral_blocks[i, 1]))
            wave_numbers.append(WaveNumber('L', spectral_blocks[i, 1], 0, spectral_blocks[i, 0], spectral_blocks[i, 0], spectral_blocks[i, 1]))

    return np.array(wave_numbers)


def basin_wavenumbers(spectral_blocks):
    """Functions that returns the :class:`WaveNumber` objects corresponding to a given list of spectral blocks for a
    closed basin spatial domain.

    Parameters
    ----------
    spectral_blocks: ~numpy.ndarray(int)
        Spectral blocks detailing the modes :math:`x`- and :math:`y`-wavenumber.
        Array of shape (nblocks, 2), where `nblocks` is the number of spectral blocks.

    Returns
    -------
    ~numpy.ndarray(WaveNumber)
        The array of wavenumber objects corresponding to the given spectral blocks.
    """

    # initialization of the variables
    wave_numbers = list()

    # Oceanic wavenumbers definition

    for i in range(spectral_blocks.shape[0]):  # function type is limited to L for the moment: ocean is a closed basin
        wave_numbers.append(WaveNumber('L', spectral_blocks[i, 1], 0, spectral_blocks[i, 0], spectral_blocks[i, 0] / 2., spectral_blocks[i, 1]))

    return np.array(wave_numbers)



"""
    Plotting utility module
    =======================

    Work in progress...

"""
import matplotlib.pyplot as plt


def std_plot(x, mean, std, ax=None, **kwargs):
    """Plot a mean and a standard deviation time series.

    Parameters
    ----------
    x: ~numpy.ndarray
        1D array of the x-axis values (usually time)
    mean: ~numpy.ndarray
        1D array of mean values to represent, should have the same shape as x.

    std: ~numpy.ndarray
        1D array of standard deviation values to represent, should have the same shape as mean.
    ax: None or ~matplotlib.axes.Axes
        A `matplotlib`_ axes instance to plot the values. If `None`, create one.
    kwargs: dict
        Keyword arguments to be given to the plot routine.

    Returns
    -------
    ~matplotlib.axes.Axes
        The axes instance on which the values where plotted.


    .. _matplotlib: https://matplotlib.org/
    """

    if ax is None:
        fig = plt.figure()
        ax = fig.gca()

    ax.fill_between(x, mean - std, mean + std, **kwargs)

    return ax
